/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: pc_apreq.c
*    Creation Date: 12 Feb 2001 (v0.00)
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:54:05 $
*
****************************************************************************
*    Description:
*
*       Implementation of packetCable AP_REQ and AP_REP
*
****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#ifdef WIN32
#include <winsock2.h>
#include "haus_cpu.h"
#endif
#include "seckrb.h"
#include "SecLib.h"
#include "krb5client.h"


/*
*****************************************************************************
** FUNCTION:     apreqMakeProvASD
**
** PURPOSE:      construct the ASD pointer for AP request for Provisioning
**
** PARAMETERS:    pProvAsd - in - pointer to ASD structure
**                pAsd - out - output pointer to ASD
**                pAsdLen - out - length of output pointer
**
** RETURNS:      none


**
** NOTE:
**
**
*****************************************************************************
*/
void apreqMakeProvASD(PROV_ASD_APREQ *pProvAsd, char *pAsd, UINT32 *pAsdLen)
{
   UINT32 i=0;
   UINT32 temp;

   /* snmpEngineId Length and snmpEngineId */
   memcpy(&pAsd[i++], &(pProvAsd->engIdLen), 1);
   memcpy(&pAsd[i], pProvAsd->engId, pProvAsd->engIdLen);
   i += pProvAsd->engIdLen;

   /* snmpEngingBoots and snmpEngineTime */
#if (CPU_LITTLE_ENDIAN != 0)
   temp = htonl(pProvAsd->engBoots);
   memcpy(&pAsd[i], &temp, 4);
   i += 4;
   temp = htonl(pProvAsd->engTime);
   memcpy(&pAsd[i], &temp, 4);
   i += 4;
#else
   memcpy(&pAsd[i], &(pProvAsd->engBoots), 4);
   i += 4;
   memcpy(&pAsd[i], &(pProvAsd->engTime), 4);
   i += 4;
#endif

   /* UsmUserName Length and UsmUserName */
   memcpy(&pAsd[i++], &(pProvAsd->userNameLen), 1);
   memcpy(&pAsd[i], pProvAsd->userName, pProvAsd->userNameLen);
   i += pProvAsd->userNameLen;

   *pAsdLen = i;
}

/*
*****************************************************************************
** FUNCTION:     apreqMakeCmsASD
**
** PURPOSE:      construct the ASD pointer for AP request for Callagent
**
** PARAMETERS:    pProvAsd - in - pointer to ASD structure
**                pAsd - out - output pointer to ASD
**                pAsdLen - out - length of output pointer
**
** RETURNS:      none
**
** NOTE:
**
**
*****************************************************************************
*/
void apreqMakeCmsASD(CMS_ASD_APREQ *pProvAsd, char *pAsd, UINT32 *pAsdLen)
{
   int i = 4;
   UINT32 temp;

#if (CPU_LITTLE_ENDIAN != 0)
   temp = htonl(pProvAsd->inbound_SPI);
   memcpy(pAsd, &temp, i);
#else
   memcpy(pAsd, &(pProvAsd->inbound_SPI), i);
#endif

   *pAsdLen = i;
}

/*
*****************************************************************************
** FUNCTION:     make_pc_ap_req
**
** PURPOSE:      Add packet cable defined fields to the constructed AP_REQ
**
** PARAMETERS:   inbuf - in - the asn.1 encoded AP_REQ message
**               inkey - in - session key for signing
**               pAppData - in - ASD pointer
**               msg - out - the packet cable defined AP_REQ
**               msg_len - out - length of msg
**               app_flag - in - type of application: IPSec or SNMPv3
**               server_nonce - in - the value of sever nonce
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
int make_pc_ap_req(
      krb5_data        *inbuf,
      krb5_data        *inkey,
      void             *pAppData,
      unsigned char    *msg,
      int              *msg_len,
      const int        app_flag,
      const int        server_nonce,
      int              authTID,
      int              encTID)
{
   int               ix=0;
   unsigned char     sig[20];
   unsigned char     *tmp_msg;
   int               tmp_len;
   UINT32            temp;

   CMS_ASD_APREQ *pCmsAsd   = NULL;
   PROV_ASD_APREQ *pProvAsd = NULL;
   char *pAsd;
   UINT32 pAsdLen;
   unsigned char  ap_key[20];

   msg[ix++] = 0x02;          /* msg-id */
   msg[ix++] = app_flag;      /* DOI */
   msg[ix++] = 0x10;          /* protocol version */
   memcpy(&msg[ix], inbuf->data, inbuf->length);   /* AP_REQ */
   ix += inbuf->length;

   /* copy the value of server-nonce */
#if (CPU_LITTLE_ENDIAN != 0)
   temp = htonl(server_nonce);
   memcpy(&msg[ix], &temp, 4);
   ix += 4;
#else
   memcpy(&msg[ix], &(server_nonce), 4);
   ix += 4;
#endif

   /* ASD */
   if (app_flag == APPLICATION_PROV)
   {
      pAsd = (char *)krbAllocBuf(sizeof(PROV_ASD_APREQ));
      pProvAsd = (PROV_ASD_APREQ*) pAppData;
      apreqMakeProvASD(pProvAsd, pAsd, &pAsdLen);
   }
   else
   {
      pAsd = (char *)krbAllocBuf(sizeof(CMS_ASD_APREQ));
      pCmsAsd = (CMS_ASD_APREQ*) pAppData;
      apreqMakeCmsASD(pCmsAsd, pAsd, &pAsdLen);
   }
   memcpy(&msg[ix], pAsd, pAsdLen);
   ix += pAsdLen;
   krbFreeBuf(pAsd);

   msg[ix++] = 0x01;   /* length of ciphersuite */
   if (app_flag == APPLICATION_PROV)
   {
      msg[ix++] = SNMPV3_HMAC_MD5_TID; /* Authentication transform Id-HMAC-MD5*/
      msg[ix++] = SNMPV3_DES_TID;   /* Encryption algorithm Id - DES */
   }
   else
   {
      /* if application is CMS, see if we need to get the TIDs from
       * the rekey message */
      if ( authTID )
      {
         msg[ix++] = authTID;
      }
      else
      {
         msg[ix++] = IPSec_ESP_HMAC_MD5_TID;
      }
      if ( encTID )
      {
         msg[ix++] = encTID;
      }
      else
      {
         msg[ix++] = IPSec_ESP_3DES_TID;
      }
   }

   msg[ix++] = 0x1;    /* re-establish flag */

   /* hmac changes the data, so make a copy */
   tmp_msg = (unsigned char *)krbAllocBuf(ix);
   memcpy(tmp_msg, msg, ix);

   SecLibGetCryptoIface()->secCryptoCalcHash(HASH_SHA1_TYPE, inkey->data,
         inkey->length, NULL, 0, ap_key ,&tmp_len);

   /* Sign the message */
   /* Call to SecLib to do HMAC_SHA1 */
   SecLibGetCryptoIface()->secCryptoCalcHash(HASH_SHA1_TYPE, tmp_msg, ix, ap_key,
         20, sig, &tmp_len);

   krbFreeBuf(tmp_msg);
   memcpy(&msg[ix], sig, 20);
   ix += 20;

   *msg_len = ix;

   return 1;
}

