/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: asreq.c
*    Creation Date: 27 Oct 2000 (v0.00)
*    VSS Info:
*        $Revision: 1.13 $
*        $Date: 2005/08/24 20:25:38 $
*
****************************************************************************
*    Description:
*
*      Implementation of packetCable PKINIT Request and reply
*
****************************************************************************/

 
///#include <xbcmos.h>
#include "bcmosHelper.h"
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "seckrb.h"
#include "krb5client.h"
#include "pkinit.h"
#include "pkinit_checksum.h"
#include "krbasn1.h"
#include "SecLib.h"
#include <SecUtils.h>
#include "KerbSnmpConfig.h"

/* Diffie hellman oakley group 1 prime and generator */
const unsigned char OAKLEY1_PRIME[] =
{
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xC9, 0x0F, 0xDA, 0xA2,
    0x21, 0x68, 0xC2, 0x34, 0xC4, 0xC6, 0x62, 0x8B, 0x80, 0xDC, 0x1C, 0xD1,
    0x29, 0x02, 0x4E, 0x08, 0x8A, 0x67, 0xCC, 0x74, 0x02, 0x0B, 0xBE, 0xA6,
    0x3B, 0x13, 0x9B, 0x22, 0x51, 0x4A, 0x08, 0x79, 0x8E, 0x34, 0x04, 0xDD,
    0xEF, 0x95, 0x19, 0xB3, 0xCD, 0x3A, 0x43, 0x1B, 0x30, 0x2B, 0x0A, 0x6D,
    0xF2, 0x5F, 0x14, 0x37, 0x4F, 0xE1, 0x35, 0x6D, 0x6D, 0x51, 0xC2, 0x45,
    0xE4, 0x85, 0xB5, 0x76, 0x62, 0x5E, 0x7E, 0xC6, 0xF4, 0x4C, 0x42, 0xE9,
    0xA6, 0x3A, 0x36, 0x20, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

/* Diffie hellman oakley group 1 prime and generator */
const unsigned char OAKLEY1_Q[] =
{
    0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe4, 0x87, 0xed, 0x51,
    0x10, 0xb4, 0x61, 0x1a, 0x62, 0x63, 0x31, 0x45, 0xc0, 0x6e, 0x0e, 0x68,
    0x94, 0x81, 0x27, 0x04, 0x45, 0x33, 0xe6, 0x3a, 0x01, 0x05, 0xdf, 0x53,
    0x1d, 0x89, 0xcd, 0x91, 0x28, 0xa5, 0x04, 0x3c, 0xc7, 0x1a, 0x02, 0x6e,
    0xf7, 0xca, 0x8c, 0xd9, 0xe6, 0x9d, 0x21, 0x8d, 0x98, 0x15, 0x85, 0x36,
    0xf9, 0x2f, 0x8a, 0x1b, 0xa7, 0xf0, 0x9a, 0xb6, 0xb6, 0xa8, 0xe1, 0x22,
    0xf2, 0x42, 0xda, 0xbb, 0x31, 0x2f, 0x3f, 0x63, 0x7a, 0x26, 0x21, 0x74,
    0xd3, 0x1d, 0x1b, 0x10, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

/* Diffie hellman oakley group 2 prime and generator */
const unsigned char MTA_OAKLEY2_PRIME[] =
{
    0xFF, 0xFF, 0xFF, 0xFF ,0xFF, 0xFF, 0xFF, 0xFF ,0xC9, 0x0F, 0xDA, 0xA2,
    0x21, 0x68, 0xC2, 0x34 ,0xC4, 0xC6, 0x62, 0x8B ,0x80, 0xDC, 0x1C, 0xD1,
    0x29, 0x02, 0x4E, 0x08 ,0x8A, 0x67, 0xCC, 0x74 ,0x02, 0x0B, 0xBE, 0xA6,
    0x3B, 0x13, 0x9B, 0x22 ,0x51, 0x4A, 0x08, 0x79 ,0x8E, 0x34, 0x04, 0xDD,
    0xEF, 0x95, 0x19, 0xB3 ,0xCD, 0x3A, 0x43, 0x1B ,0x30, 0x2B, 0x0A, 0x6D,
    0xF2, 0x5F, 0x14, 0x37 ,0x4F, 0xE1, 0x35, 0x6D ,0x6D, 0x51, 0xC2, 0x45,
    0xE4, 0x85, 0xB5, 0x76 ,0x62, 0x5E, 0x7E, 0xC6 ,0xF4, 0x4C, 0x42, 0xE9,
    0xA6, 0x37, 0xED, 0x6B ,0x0B, 0xFF, 0x5C, 0xB6 ,0xF4, 0x06, 0xB7, 0xED,
    0xEE, 0x38, 0x6B, 0xFB ,0x5A, 0x89, 0x9F, 0xA5 ,0xAE, 0x9F, 0x24, 0x11,
    0x7C, 0x4B, 0x1F, 0xE6 ,0x49, 0x28, 0x66, 0x51 ,0xEC, 0xE6, 0x53, 0x81,
    0xFF, 0xFF, 0xFF, 0xFF ,0xFF, 0xFF, 0xFF, 0xFF
};

/* Diffie hellman oakley group 2 prime and generator */
const unsigned char OAKLEY2_Q[] =
{
    0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xE4, 0x87, 0xED, 0x51,
    0x10, 0xB4, 0x61, 0x1A, 0x62, 0x63, 0x31, 0x45, 0xC0, 0x6E, 0x0E, 0x68,
    0x94, 0x81, 0x27, 0x04, 0x45, 0x33, 0xE6, 0x3A, 0x01, 0x05, 0xDF, 0x53,
    0x1D, 0x89, 0xCD, 0x91, 0x28, 0xA5, 0x04, 0x3C, 0xC7, 0x1A, 0x02, 0x6E,
    0xF7, 0xCA, 0x8C, 0xD9, 0xE6, 0x9D, 0x21, 0x8D, 0x98, 0x15, 0x85, 0x36,
    0xF9, 0x2F, 0x8A, 0x1B, 0xA7, 0xF0, 0x9A, 0xB6, 0xB6, 0xA8, 0xE1, 0x22,
    0xF2, 0x42, 0xDA, 0xBB, 0x31, 0x2F, 0x3F, 0x63, 0x7A, 0x26, 0x21, 0x74,
    0xD3, 0x1B, 0xF6, 0xB5, 0x85, 0xFF, 0xAE, 0x5B, 0x7A, 0x03, 0x5B, 0xF6,
    0xF7, 0x1C, 0x35, 0xFD, 0xAD, 0x44, 0xCF, 0xD2, 0xD7, 0x4F, 0x92, 0x08,
    0xBE, 0x25, 0x8F, 0xF3, 0x24, 0x94, 0x33, 0x28, 0xF6, 0x73, 0x29, 0xC0,
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};


/* MTA device certificate */
static unsigned char    gmta_dev_cert[MAX_CERTIFICATE_LENGTH];
static int              gmta_dev_cert_len;

/* MTA manufacture certificate */
static unsigned char    gmta_man_cert[MAX_CERTIFICATE_LENGTH];
static int              gmta_man_cert_len;

/* Telephony Root Certificate for validating TGS certificate */
unsigned char          gtelrootcert[MAX_CERTIFICATE_LENGTH];
int                    gtelrootcert_len;

/* RSA key and modulous, generated at manufacturing */
static unsigned char   grsa_priv_exp[MAX_CERT_KEY_LENGTH];
int                    grsa_priv_exp_len;
static unsigned char   grsa_mod[MAX_CERT_KEY_LENGTH];
int                    grsa_mod_len;

krb5_error_code decode_krb5_pkinit_rep(
      const krb5_data * code,
      pkinit_reply **rep);

krb5_error_code decode_krb5_auth_pack(
      const krb5_data * code,
      encapsulated_content_info **rep);

krb5_error_code encode_krb5_pkinit_req(
      const pkinit_request * rep,
      krb5_data ** code);

////////////////////////////////////////////////////////////////////
/*
******************************************************************************
** FUNCTION:   secKrbSetCert_ch
**
** PURPOSE:    read the certificates and RSA keys necessary from NVRAM
**
**   PARAMETERS:
**
** RETURNS:   Success or Failure
**
** NOTE:
**
******************************************************************************
*/
#define MAX_KEY_LEN 257

int secKrbSetCert_ch(void)
{
    // Read the certificates from NVRAM and copy them into the globals given below.
    // Will need to parse private key

    //We are calling into the app here

    int           ManCertLen = 1000;  //set to max buf size
    unsigned char ManCert[1000];
    int           PSCertLen = 1000;
    unsigned char PSCert[1000];
    int           ServProvCertLen = 1000;
    unsigned char ServProvCert[1000];
    unsigned int  PsDevPrivExpLen = MAX_KEY_LEN;
    unsigned char PsDevPrivExp[MAX_KEY_LEN];
    unsigned int  PsDevKeyModLen = MAX_KEY_LEN;
    unsigned char PsDevKeyMod[MAX_KEY_LEN];

    gmta_dev_cert_len = 0;
    gmta_man_cert_len = 0;
    gtelrootcert_len  = 0;

    //printf("STARTING secKrbSetCert_ch!!!!!\n");
    if ( ! AppCHGetCertPSPrivateKeyParts(&PsDevPrivExp, &PsDevPrivExpLen, &PsDevKeyMod, &PsDevKeyModLen) )
    //if ( ! AppCHGetCertPSPrivateKeyPartsProp(&PsDevPrivExp, &PsDevPrivExpLen, &PsDevKeyMod, &PsDevKeyModLen) )
    {
        if ( ktracelevel > KTRACE_INFO )
        {
            DumpBuf("from krb5client -PRIVATE EXP",PsDevPrivExp, PsDevPrivExpLen, 'x');
            DumpBuf("from krb5client -DEV MODULUS",PsDevKeyMod,  PsDevKeyModLen, 'x');
        }
        memcpy(grsa_mod, PsDevKeyMod, PsDevKeyModLen);
        grsa_mod_len = PsDevKeyModLen;
        memcpy(grsa_priv_exp, PsDevPrivExp, PsDevPrivExpLen);
        grsa_priv_exp_len = PsDevPrivExpLen;
    }
    else
        return 1;


    if ( ! AppCHGetCertPS(&PSCert, &PSCertLen) )
    {
        //DumpBuf("PSCert ",PSCert, PSCertLen, 'x');
        if ( ktracelevel > KTRACE_INFO )
            DumpBuf("PSCert (first 30 bytes ...)",PSCert, 30, 'x');
        gmta_dev_cert_len = PSCertLen;
        memcpy(gmta_dev_cert,PSCert,PSCertLen);
    }
    else
        return 1;


    if ( ! AppCHGetCertManCA(&ManCert, &ManCertLen) )
    {
        //DumpBuf("ManCert ",ManCert, ManCertLen, 'x');
        if ( ktracelevel > KTRACE_INFO )
            DumpBuf("ManCert (first 30 bytes ...)",ManCert, 30, 'x');
        memcpy(gmta_man_cert, ManCert,ManCertLen);
        gmta_man_cert_len = ManCertLen;
    }
    else
        return 1;

    if ( ! AppCHGetCertServProvCA( &ServProvCert, &ServProvCertLen ) )
    {
        //DumpBuf("ServProvCert ",ServProvCert, ServProvCertLen, 'x');
        if ( ktracelevel > KTRACE_INFO )
            DumpBuf("ServProvCert first 30 bytes ...)",ServProvCert, 30, 'x');
        memcpy(gtelrootcert, ServProvCert, ServProvCertLen);
        gtelrootcert_len =   ServProvCertLen;
    }
    else
    {
        LIBDEBUG("AppCHGerCertServProvCA FAILED - returning error\n");
        return 1;
    }

    //LOG(("Writing certificates/Keys to NVRAM. - Cablehome certs\n"));
    if ( (grsa_mod_len!=128) && (grsa_mod_len!=256) && (grsa_mod_len!=192) )
    {
        LOG(("Device Modulus Key Length is not correct."));
        return 1;
    }
    if ( (grsa_priv_exp_len!=128)&&(grsa_priv_exp_len!=256)&&(grsa_priv_exp_len!=192) )
    {
        LOG(("Device Private Key Length is not correct."));
        return 1;
    }
    //printf("ENDING secKrbSetCert_ch!!!!!\n");
    if ( gmta_dev_cert_len &&
         gmta_man_cert_len &&
         gtelrootcert_len )
        return 0;
    return 1;

}
///////////////////////////////////////////////////////
#ifndef DEMO
/*
******************************************************************************
** FUNCTION:   secKrbSetCert
**
** PURPOSE:    read the certificates and RSA keys necessary from NVRAM
**
**   PARAMETERS:
**
** RETURNS:   Success or Failure
**
** NOTE:
**
******************************************************************************
*/
int secKrbSetCert(void)
{
   SecLibGetNVRAMIface()->secReadNVRAMMTACertKeyData(
         MTA_DEV_CERT,&gmta_dev_cert_len,gmta_dev_cert);

   SecLibGetNVRAMIface()->secReadNVRAMMTACertKeyData(
         MTA_MAN_CERT,&gmta_man_cert_len,gmta_man_cert);

   if ( cmNvGetUseNVRAMTelRootCert() )
   {
      /* clear and override telephony root cert with what's in nvram */
      SEC_LOG(("NVOL UseNVRAMTelRootCert is TRUE, Read Telephony Root Cert from NVRAM."));
      SecLibGetNVRAMIface()->secReadNVRAMMTACertKeyData(TEL_ROOT_CERT,&gtelrootcert_len,gtelrootcert);
   }
   else
   {
      if ( !sizeof(image_gtelrootcert) )
      {
         SEC_LOG(("ERROR: Telephony Root Cert from Software Load is EMPTY."));
         return 0;
      }

      if ( sizeof(image_gtelrootcert) > MAX_CERTIFICATE_LENGTH )
      {
         LOG(("ERROR: Telephony Root Cert from Software Load(%d) is bigger than 4K!.",
                  sizeof(image_gtelrootcert)));
         return 0;
      }
      LOG(("NVOL UseNVRAMTelRootCert is FALSE, Read Telephony Root Cert from SW image (%d) bytes.", sizeof(image_gtelrootcert)));
      memset(gtelrootcert, 0, MAX_CERTIFICATE_LENGTH);
      gtelrootcert_len = sizeof(image_gtelrootcert);
      memcpy(gtelrootcert, image_gtelrootcert, gtelrootcert_len);
   }

   SecLibGetNVRAMIface()->secReadNVRAMMTACertKeyData(MAT_DEV_MOD_KEY,&grsa_mod_len,grsa_mod);
   SecLibGetNVRAMIface()->secReadNVRAMMTACertKeyData(MTA_DEV_PRIV_KEY,&grsa_priv_exp_len,grsa_priv_exp);

   if ((grsa_mod_len!=128) && (grsa_mod_len!=256) && (grsa_mod_len!=192))
   {
        LOG(("Device Modulus Key Length is not correct."));
      return 1;
   }
   if ((grsa_priv_exp_len!=128)&&(grsa_priv_exp_len!=256)&&(grsa_priv_exp_len!=192))
   {
        LOG(("Device Private Key Length is not correct."));
      return 1;
   }

   if ( gmta_dev_cert_len &&
         gmta_man_cert_len &&
         gtelrootcert_len )
      return 0;
   return 1;
}
#else
/*
******************************************************************************
** FUNCTION:   secKrbSetCert
**
** PURPOSE:    read the certificates and RSA keys necessary from NVRAM
**
**   PARAMETERS:
**    1. pMTADevCert - MTA device certificate
**    2. pMTADevCertLen - length of MTA device certificate
**    3. pMTAManCert - MTA man certificate
**    4. pMTAManCertLen - length of MTA man certificate
**    5. pMTARsaMod - MTA's RSA Moduolus
**    6. pMTARsaModLen - MTA's RSA Modulous length
**    7. pMTARsaPrivExp - MTA's RSA private exponent
**    8. pMTARsaPrivExpLen - MTA's RSA private exponent length
**    9. pServProvCert - Service Provider's certificate
**    10. pServProvCertLen - Service Provider certificate length
**
**
** RETURNS:   Success or Failure
**
** NOTE:
**
******************************************************************************
*/
int secKrbSetCert(
                 const unsigned char *pMTADevCert,
                 const int pMTADevCertLen,
                 const unsigned char *pMTAManCert,
                 const int pMTAManCertLen,
                 const unsigned char *pMTARsaMod,
                 const int pMTARsaModLen,
                 const unsigned char *pMTARsaPrivExp,
                 const int pMTARsaPrivExpLen,
                 const unsigned char *pServProvCert,
                 const int pServProvCertLen)
{

    ch_secReadNVRAMMTACertKeyData(MTA_DEV_CERT,&gmta_dev_cert_len,gmta_dev_cert);
    ch_secReadNVRAMMTACertKeyData(MTA_MAN_CERT,&gmta_man_cert_len,gmta_man_cert);
    ch_secReadNVRAMMTACertKeyData(TEL_ROOT_CERT,&gtelrootcert_len,gtelrootcert);
    ch_secReadNVRAMMTACertKeyData(MAT_DEV_MOD_KEY,&grsa_mod_len,grsa_mod);
    ch_secReadNVRAMMTACertKeyData(MTA_DEV_PRIV_KEY,&grsa_priv_exp_len,grsa_priv_exp);

    return 1;
}
#endif

/*
*****************************************************************************
** FUNCTION:     prep_pkinit_req
**
** PURPOSE:      put a pkinit request structure together
**
** PARAMETERS:   request - in - kdc-req-body to be used for checksum
**               rpk - the pkinit request structure
**               pkinit_nonce - out: nonce generated
**               mtaprivkey - our: provate ket generated for mta
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
krb5_error_code prep_pkinit_req(
      krb5_kdc_req         *request,
      pkinit_request       *rpk,
      int                  *pkinit_nonce,
      priv_key_t           *mtaprivkey,
      int                  app_flag)
{
   unsigned char     sernum[32];
   unsigned long     sernumlen;
   krb5_data         *scratch;
   krb5_data         *scratch2;
   RSA_key           *rsa_key;
   int               i;
   int               checksum_len;
   int               status;
   int               clock_skew;
   int               tmp;
   int               utc_time_offset;
   int               sec_debug;
   RSA_key           *tmpkey;
   int               pubKeyId;

   /* time related declarations */
   OS_TMDATE   idate;
   OS_TMTIME   itime;
   OS_TICKS    iticks;
   OS_STATUS   istat;
   time_t      tmp_time;
   int         tmp_time_ms;

   krb5_error_code ret;
    
   rpk->signed_authpack.content_type.oid_array[0] = 1;
   rpk->signed_authpack.content_type.oid_array[1] = 2;
   rpk->signed_authpack.content_type.oid_array[2] = 840;
   rpk->signed_authpack.content_type.oid_array[3] = 113549;
   rpk->signed_authpack.content_type.oid_array[4] = 1;
   rpk->signed_authpack.content_type.oid_array[5] = 7;
   rpk->signed_authpack.content_type.oid_array[6] = 2;
   rpk->signed_authpack.content_type.length = 7;

   rpk->signed_authpack.signed_data.cms_version = 3;

   rpk->signed_authpack.signed_data.digest_alg_ids.digest_alg_id.alg_oid.oid_array[0] = 1;
   rpk->signed_authpack.signed_data.digest_alg_ids.digest_alg_id.alg_oid.oid_array[1] = 3;
   rpk->signed_authpack.signed_data.digest_alg_ids.digest_alg_id.alg_oid.oid_array[2] = 14;
   rpk->signed_authpack.signed_data.digest_alg_ids.digest_alg_id.alg_oid.oid_array[3] = 3;
   rpk->signed_authpack.signed_data.digest_alg_ids.digest_alg_id.alg_oid.oid_array[4] = 2;
   rpk->signed_authpack.signed_data.digest_alg_ids.digest_alg_id.alg_oid.oid_array[5] = 26;
   rpk->signed_authpack.signed_data.digest_alg_ids.digest_alg_id.alg_oid.length = 6;
   rpk->signed_authpack.signed_data.digest_alg_ids.digest_alg_id.parameter.type = 0;
   rpk->signed_authpack.signed_data.digest_alg_ids.length = 1;


   /* setting ecnapsulatedContentInfo:content_Type */
   rpk->signed_authpack.signed_data.encap_content_info.econtent_type.oid_array[0] = 1;
   rpk->signed_authpack.signed_data.encap_content_info.econtent_type.oid_array[1] = 3;
   rpk->signed_authpack.signed_data.encap_content_info.econtent_type.oid_array[2] = 6;
   rpk->signed_authpack.signed_data.encap_content_info.econtent_type.oid_array[3] = 1;
   rpk->signed_authpack.signed_data.encap_content_info.econtent_type.oid_array[4] = 5;
   rpk->signed_authpack.signed_data.encap_content_info.econtent_type.oid_array[5] = 2;
   rpk->signed_authpack.signed_data.encap_content_info.econtent_type.oid_array[6] = 3;
   rpk->signed_authpack.signed_data.encap_content_info.econtent_type.oid_array[7] = 1;
   rpk->signed_authpack.signed_data.encap_content_info.econtent_type.length=8;

    rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.nonce = rand();
    LIBDEBUG("pkinit pkauthenticator nonce = %d.\n",
               rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.nonce);
    *pkinit_nonce =
    rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.nonce;

    /* get current time */
    //trc-  use  TimeMs() from bcmos
    //istat = bcmOsGetTimeEx(&idate, &itime, &iticks);
    //if (OS_STATUS_OK != istat)
    //{
    //   printf("Could not get the time. error code is %d\n", istat);
    //   return 0;
    //}

   if ( app_flag == APPLICATION_CMS )
   {
        clock_skew =  ch_secReadNVRAMSkew(AS_CMS_FLAG,request->client->realm.data,
                                                      request->client->realm.length);
   }
   else
   {
        clock_skew =  ch_secReadNVRAMSkew(AS_PROV_FLAG,request->client->realm.data,
                                                       request->client->realm.length);
   }
  
    //printf("trc forcing skew !!!!!!!!\n"); 
    //clock_skew = 20000;
    LIBDEBUG("AS Clock Skew for realm %s = %d.\n", request->client->realm.data,clock_skew);

    time(&tmp_time);
#ifdef DEMO
    utc_time_offset = 0;
#else
   utc_time_offset = cmMiscGetUTCTimeOffset();
#endif
   if ( (utc_time_offset < -86400) || (utc_time_offset > 86400) )
      utc_time_offset = 0;
    LOG(("CM UTC Time Offset is %d.", utc_time_offset));
    tmp_time -= utc_time_offset;


    rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.c_time = tmp_time - clock_skew;
    LIBDEBUG("pkinit time is %d.\n",
               rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.c_time);
    //rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.cu_sec = (iticks*OS_MSEC_PER_TICK*1000)%1000;
    rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.cu_sec = (tmp_time_ms%1000)*1000;


    //trc- use timeMs from bcmos
    //printf("TICKVALUE -trc\n");
    //iticks = (tickGet());
    //printf("tickGet = (%d)\n",iticks);
    //iticks = (tickGet()*1*1000)%1000;
    iticks= 0;
    //printf("tickGet after calc = (%d)\n",iticks);


    rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.cu_sec =  iticks;//(tickGet()*10*1000)%1000;//(TimeMs());

    //rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.cu_sec =  (tickGet()*10*1000)%1000;//(TimeMs());

    ret= do_pkinit_pachecksum(request,
                              &(rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.pa_checksum.contents),
                              &checksum_len);

    //printf("\n.k  Returned from do_pkinit_pachecksum with a krb5_error_code (%d)\n\n",ret);

   rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.pa_checksum.checksum_type =
      CKSUMTYPE_RSA_MD5;
   rpk->signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.pa_checksum.length = checksum_len;

   /* setting diffie-hellman parameters */
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.oid_array[0] = 1;
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.oid_array[1] = 2;
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.oid_array[2] = 840;
#if 1
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.oid_array[3] = 10046;
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.oid_array[4] = 2;
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.oid_array[5] = 1;
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.length = 6;
#endif
#if 0
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.oid_array[3] = 113549;
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.oid_array[4] = 1;
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.oid_array[5] = 3;
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.oid_array[6] = 1;
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dhkey_oid.length = 7;
#endif

   /* The prime needs to be encoded as integer, Since the first 2 bytes are FF FF,
    * it means that 0x00 needs to be prepended to the prime */
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dh_parameter.prime = krbAllocBuf(MTA_OAKLEY2_PRIME_LEN+1);
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dh_parameter.prime[0] = 0x00;
   memcpy(rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dh_parameter.prime+1,
         MTA_OAKLEY2_PRIME, MTA_OAKLEY2_PRIME_LEN);
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dh_parameter.prime_len = MTA_OAKLEY2_PRIME_LEN+1;

   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dh_parameter.q = krbAllocBuf(OAKLEY2_Q_LEN);
   memcpy(rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dh_parameter.q,
           OAKLEY2_Q, OAKLEY2_Q_LEN);
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dh_parameter.q_len = OAKLEY2_Q_LEN;


   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dh_parameter.base = OAKLEY2_BASE;
   rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dh_parameter.private_len = 0;     /* optional */

    /////////////////////////////////////////////
    // old SecLibEcosDemo.a
    // SecLibGetCryptoIface()->secCryptoDhGener(MTA_OAKLEY2_PRIME, MTA_OAKLEY2_PRIME_LEN, OAKLEY2_BASE,
    //                                         &(rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_pub_key),
    //                                         &(rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_pub_key_len),
    //                                         &(mtaprivkey->priv_key), &(mtaprivkey->priv_key_len));
    
    // allocate memory for the public and private keys.  Where does it get freed? Is it a leak?  I don't know
    // but this makes it the same as using the older version of secCryptoDhGener which did a malloc.
    rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_pub_key = (unsigned char *)malloc(256);
    mtaprivkey->priv_key = (unsigned char*)malloc(256);

    SecLibGetCryptoIface()->secCryptoDhGener(MTA_OAKLEY2_PRIME, MTA_OAKLEY2_PRIME_LEN, OAKLEY2_BASE,
                                            (rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_pub_key),
                                            &(rpk->signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_pub_key_len),
                                            (mtaprivkey->priv_key), &(mtaprivkey->priv_key_len));
    ////////////////////////////////////////////

   /* add the MTA device certificate */
   rpk->signed_authpack.signed_data.certificates.certs[0].cert_len = gmta_dev_cert_len;
   rpk->signed_authpack.signed_data.certificates.certs[0].cert = gmta_dev_cert;

   /* Add the MTA Munufacture certificate */
   rpk->signed_authpack.signed_data.certificates.certs[1].cert_len = gmta_man_cert_len;
   rpk->signed_authpack.signed_data.certificates.certs[1].cert = gmta_man_cert;

   rpk->signed_authpack.signed_data.certificates.length = 2;

   /* signerInfos */
   rpk->signed_authpack.signed_data.signer_infos.signer_info.cms_version = 1;

   if ( 0 == (SecLibGetCertIface()->secCertGetSerNum(gmta_dev_cert,gmta_dev_cert_len, sernum, &sernumlen) ) )
   {
   rpk->signed_authpack.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_sernum = krbAllocBuf(sernumlen+1);
   memcpy(rpk->signed_authpack.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_sernum, sernum, sernumlen);
   rpk->signed_authpack.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_sernum_len = sernumlen;
       if ( ktracelevel > KTRACE_INFO )
       {
           DumpBuf("Serial Number from DEV cert",
                   (rpk->signed_authpack.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_sernum),
                   (rpk->signed_authpack.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_sernum_len),
                   'x');
       }
   }
   else
   {
       LIBDEBUG("ERROR - secCertGetSerNum returned with problem\n");
   }
   
  if ( (SecLibGetCertIface()->secCertGetIssuerNameOS(gmta_dev_cert, gmta_dev_cert_len,
                                                      &(rpk->signed_authpack.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_name),
                                                      &(rpk->signed_authpack.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_name_len) )) )
       printf("ERROR -secCertGetIssuerNamOS ret with problem\n");

   if ( ktracelevel > KTRACE_INFO )
   {
        DumpBuf("signer name",
                (rpk->signed_authpack.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_name),
                (rpk->signed_authpack.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_name_len),
                'x');
    }

   /* SHA-1 in Digest_alg */
   rpk->signed_authpack.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.length = 6;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.oid_array[0] = 1;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.oid_array[1] = 3;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.oid_array[2] = 14;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.oid_array[3] = 3;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.oid_array[4] = 2;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.oid_array[5] = 26;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.digest_alg_id.parameter.type = 0;

   /* RSA over SHA-1 oid in signature_alg */
   rpk->signed_authpack.signed_data.signer_infos.signer_info.signature_alg_id.alg_oid.length = 7;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.signature_alg_id.alg_oid.oid_array[0] = 1;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.signature_alg_id.alg_oid.oid_array[1] = 2;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.signature_alg_id.alg_oid.oid_array[2] = 840;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.signature_alg_id.alg_oid.oid_array[3] = 113549;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.signature_alg_id.alg_oid.oid_array[4] = 1;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.signature_alg_id.alg_oid.oid_array[5] = 1;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.signature_alg_id.alg_oid.oid_array[6] = 1;
   rpk->signed_authpack.signed_data.signer_infos.signer_info.signature_alg_id.parameter.type = 0;

   /* sign and add signature */
   scratch = (krb5_data *)krbAllocBuf(sizeof(krb5_data));
   if (scratch == NULL)
   {
      return ENOMEM;
   }
   encode_krb5_pkinit_authpack(&rpk->signed_authpack.signed_data.encap_content_info.econtent, &scratch);

   ALLOC_RSA_KEY(tmpkey);
   if ( SecLibGetCertIface()->secCertGetPublicKey( 
            gmta_dev_cert,
            gmta_dev_cert_len, 
            tmpkey, 
            &pubKeyId ) )
   {
      printf("ERROR While getting the public key of CA(KDC) cert.\n");
      return  KRB5_CONFIG_BADFORMAT;
   }

   ALLOC_RSA_KEY(rsa_key);
   CREATE_RSA_KEY(
         rsa_key, 
         grsa_mod, 
         grsa_priv_exp, 
         tmpkey->exp,
         grsa_mod_len, 
         grsa_priv_exp_len,
         tmpkey->expBufLen);

   status = SecLibGetCryptoIface()->secCryptoSign(scratch->data,scratch->length,rsa_key,
         &(rpk->signed_authpack.signed_data.signer_infos.signer_info.signature.signature_val),
         &(rpk->signed_authpack.signed_data.signer_infos.signer_info.signature.signature_len));
    //trc  free the key
    //FREE_RSA_KEY( rsa_key,0 );

   if ( status )
   {
      LIBDEBUG("ERROR: Signature Process failed errorcode = %d.\n", status);
      return(status);
   }

   FREE_RSA_KEY( rsa_key,0 );
   FREE_RSA_KEY( tmpkey,0 );
   krbFreeBuf(scratch->data);
   krbFreeBuf(scratch);

   return 0;
}


/*
*****************************************************************************
** FUNCTION:     prep_provserv_preauth
**
** PURPOSE:      put together a PROC-SRV-LOCATION preauthenticator
**
** PARAMETERS:   prov_serv_fqdn - in - prov server fqdn
**               ps_preauth - out: encoded pre-authenticator
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
krb5_error_code prep_provserv_preauth(
                                     const char *prov_serv_fqdn,
                                     krb5_data **ps_preauth)
{
    krb5_data *ps_pa;
    asn1_error_code  retval;

    ps_pa = (krb5_data *)krbAllocBuf(sizeof(krb5_data));
    if ( ps_pa == NULL )
    {
        retval = ENOMEM;
        return(retval);
    }

    ps_pa->length = 12;//strlen(prov_serv_fqdn);
    ps_pa->data = (char *)krbAllocBuf(ps_pa->length);
    memcpy(ps_pa->data, "doweneedthis", ps_pa->length);
    printf("putting doweneedthis in the prauth for provserv\n");

    //ps_pa->length = strlen(prov_serv_fqdn);
    //ps_pa->data = (char *)krbAllocBuf(ps_pa->length);
    //memcpy(ps_pa->data, prov_serv_fqdn, ps_pa->length);

    retval = encode_krb5_provserv(ps_pa, ps_preauth);
    if ( retval )
    {
        LIBDEBUG("Problems encoding prov server preauth. error code = %d.\n",retval);
        krbFreeBuf(ps_pa->data);
        krbFreeBuf(ps_pa);
        return(retval);
    }
    return 0;
}

/*
*****************************************************************************
** FUNCTION:     make_krb5_pkinit_req
**
** PURPOSE:      encodes a pkinit request (ASN.1)
**
** PARAMETERS:   request - in - AS request to be encoded for checksum
**               out_padata - out: encoded pkinit request
**               pkinit_nonce - out: nonce generated
**               mtaprivkey - out: private key generated
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
krb5_error_code make_krb5_preauth(
                                 krb5_kdc_req *request,
                                 const char *prov_serv_fqdn,
                                 int *pkinit_nonce,
                                 priv_key_t *mtaprivkey,
                                 int  app_flag)
{
    krb5_pa_data     *pkinit_pa;
    //krb5_pa_data     *prov_serv_pa;
    krb5_data        *scratch;
    pkinit_request   in_pkinit;
    asn1_error_code  retval;

    /* make and add pkinit preauth */
    if ( ( retval = prep_pkinit_req(request, &in_pkinit, pkinit_nonce, mtaprivkey, app_flag)) != 0 )
    {
        LIBDEBUG("PKINIT preparation failed.\n");
        goto cleanup;
    }
    scratch = (krb5_data *)krbAllocBuf(sizeof(krb5_data));
    if ( scratch == NULL )
    {
        retval = ENOMEM;
        goto cleanup;
    }

    if ( (retval = encode_krb5_pkinit_req(&in_pkinit, &scratch)) != 0 )
    {
        LIBDEBUG("Problems encoding pkinit request error code %d.\n", retval);
        krbFreeBuf(scratch);
        goto cleanup;
    }

    if ( (pkinit_pa = (krb5_pa_data *)krbAllocBuf(sizeof(krb5_pa_data))) == NULL )
    {
        retval = ENOMEM;
        krbFreeBuf(scratch);
        goto cleanup;
    }

    pkinit_pa->magic = KV5M_PA_DATA;
    pkinit_pa->pa_type = KRB5_PADATA_PA_PK_AS_REQ;
    pkinit_pa->length = scratch->length;
    pkinit_pa->contents = (krb5_octet *)krbAllocBuf(pkinit_pa->length);
    if ( pkinit_pa->contents == NULL )
    {
        retval = ENOMEM;
        krbFreeBuf(scratch);
        goto cleanup;
    }
    memcpy(pkinit_pa->contents, scratch->data, pkinit_pa->length);
    krbFreeBuf(scratch->data);
    krbFreeBuf(scratch);

    /* now make and add prov-serv-location preauth */
    scratch = (krb5_data *)krbAllocBuf(sizeof(krb5_data));
    if ( scratch == NULL )
    {
        retval = ENOMEM;
        goto cleanup;
    }
#if 0
    //For CableHome we don't need pre-authenticator for provServer location
    //per CH spec section 11.3.1.1
    retval = prep_provserv_preauth(prov_serv_fqdn, &scratch);
    if ( retval )
    {
        printf("Problems with prov server preauth. error code %d.\n", retval);
        krbFreeBuf(scratch);
        goto cleanup;
    }

    if ( (prov_serv_pa = (krb5_pa_data *)krbAllocBuf(sizeof(krb5_pa_data))) == NULL )
    {
        retval = ENOMEM;
        krbFreeBuf(pkinit_pa->contents);
        krbFreeBuf(pkinit_pa);
        krbFreeBuf(scratch->data);
        krbFreeBuf(scratch);
        goto cleanup;
    }
    prov_serv_pa->pa_type = -1;
    prov_serv_pa->length = scratch->length;
    prov_serv_pa->contents = (krb5_octet *)krbAllocBuf(prov_serv_pa->length);
    if ( prov_serv_pa->contents == NULL )
    {
        retval = ENOMEM;
        goto cleanup;
    }
    memcpy(prov_serv_pa->contents, scratch->data, prov_serv_pa->length);
    krbFreeBuf(scratch->data);
    krbFreeBuf(scratch);
#endif
    request->padata[0] = pkinit_pa;
    //printf("PUTTING A NULL in for prov_serv_pa\n");
    //request->padata[1] = prov_serv_pa;
    request->padata[1] = NULL;
    request->padata[2] = NULL;

    cleanup:
    krbFreeBuf(in_pkinit.signed_authpack.signed_data.encap_content_info.econtent.pk_authenticator.pa_checksum.contents);
    krbFreeBuf(in_pkinit.signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_alg.dh_parameter.prime);
    SEC_FREE(in_pkinit.signed_authpack.signed_data.encap_content_info.econtent.client_pub_value.subj_pub_key);
    SEC_FREE(in_pkinit.signed_authpack.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_name);
    SEC_FREE(in_pkinit.signed_authpack.signed_data.signer_infos.signer_info.signature.signature_val);
    
    return retval;
}


/*
*****************************************************************************
** FUNCTION:     get_krb5_pkinit_rep
**
** PURPOSE:      decode a pkinit reply (ASN.1)
**
** PARAMETERS:   in_padata - encoded pkinit reply
**               out_pk - out: pkinit reply structure
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
krb5_error_code get_krb5_pkinit_rep(krb5_pa_data *in_padata,
                                    pkinit_reply **out_pk)
{
    krb5_data       scratch;
    asn1_error_code retval;

    scratch.data = (krb5_octet *)krbAllocBuf(in_padata->length);
    if ( scratch.data == NULL )
    {
        return ENOMEM;
    }
    memcpy(scratch.data, in_padata->contents, in_padata->length);
    scratch.length = in_padata->length;

    if ( (retval = decode_krb5_pkinit_rep(&scratch, out_pk)) != 0 )
    {
        krbFreeBuf(scratch.data);
        return retval;
    }

    krbFreeBuf(scratch.data);

    return 0;
}

/*
*****************************************************************************
** FUNCTION:     get_krb5_auth_pack_reply
**
** PURPOSE:      decode econtent of as_reply (ASN.1)
**
** PARAMETERS:   io_authpack - out: auth_pack structure decoded
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
krb5_error_code get_krb5_auth_pack_reply(encapsulated_content_info_reply *io_authpack)
{
    krb5_data       scratch;
    asn1_error_code retval;

    scratch.length = io_authpack->length;
    scratch.data = (krb5_octet *)krbAllocBuf(io_authpack->length);
    if ( scratch.data == NULL )
    {
        return ENOMEM;
    }

    memcpy(scratch.data, io_authpack->contents, io_authpack->length);

    if ( (retval = decode_krb5_auth_pack_reply(&scratch, &io_authpack)) != 0 )
        return retval;

    krbFreeBuf(scratch.data);

    return 0;


}
