/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: pkinit_checksum.c
*    Creation Date: 16 Jan 2001 (v0.00)
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:54:06 $
*
****************************************************************************
*    Description:
*
*       Implementation of packetCable PKINIT pachecksum
*
****************************************************************************/

#include "pkinit_checksum.h"



/*
*****************************************************************************
** FUNCTION:     do_pkinit_pachecksum
**
** PURPOSE:      calculate pachecksum for the pkinit request
**
** PARAMETERS:   request - in - the AS request, to do the checksum over
**               cs - out - calculated checksum
**               csLen - out - length of resulting checksum
**
** RETURNS:      success or error code
**
** NOTE:
**      
**
*****************************************************************************
*/

krb5_error_code do_pkinit_pachecksum(
      krb5_kdc_req *request, 
      krb5_octet **out_cksum, int *out_cksum_len)
{
   krb5_data          *req_data;
   krb5_checksum      cksum;
   int                sumlen;
   krb5_error_code    code;

   /* initialize the the cksum */
   if (code = krb5_c_checksum_length(CKSUMTYPE_RSA_MD5, &sumlen))
       return(code);

   cksum.checksum_type = CKSUMTYPE_RSA_MD5;
   cksum.length = sumlen;
 
   /* encode kdc-req-body to be passed to checksum */
   if ((code = encode_krb5_kdc_req_body(request, &req_data)))
   {
      return(code);
   }

   if (code = krb5_c_make_checksum(CKSUMTYPE_RSA_MD5, 0, 0,
				   req_data, &cksum)) 
   {
      return(code);
   }

   krbFreeBuf((char *)req_data->data);
   krbFreeBuf ((char *)req_data);


   *out_cksum = cksum.contents;
   *out_cksum_len = cksum.length;

   return(0);
}
