/***************************************************************************
 *    Copyright 2000  Broadcom Corporation
 *    All Rights Reserved
 *    No portions of this material may be reproduced in any form without the
 *    written permission of:
 *             Broadcom Corporation
 *             16251 Laguna Canyon Road
 *             Irvine, California  92618
 *    All information contained in this document is Broadcom Corporation
 *    company private, proprietary, and trade secret.
 *
 ****************************************************************************
 *
 *    Filename: sendto_server.c
 *    Creation Date: 15 June 2001 (v0.00)
 *    VSS Info:
 *        $Revision: 1.8 $
 *        $Date: 2004/10/04 19:52:17 $
 *
 ****************************************************************************
 *    Description:
 *
 *       Implementation of UDP send to servers (KDC, PS, CMS)
 *
 ****************************************************************************/

/*
 * (Broadcom: sendto_server.c - send to CMS, PS or KDC)
 *
 * Copyright 1990,1991 by the Massachusetts Institute of Technology.
 * All Rights Reserved.
 *
 * Export of this software from the United States of America may
 *   require a specific license from the United States Government.
 *   It is the responsibility of any person or organization contemplating
 *   export to obtain such a license before exporting.
 *
 * WITHIN THAT CONSTRAINT, permission to use, copy, modify, and
 * distribute this software and its documentation for any purpose and
 * without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that
 * the name of M.I.T. not be used in advertising or publicity pertaining
 * to distribution of the software without specific, written prior
 * permission.  Furthermore if you modify this software you must label
 * your software as modified software and not distribute it in such a
 * fashion that it might be confused with the original M.I.T. software.
 * M.I.T. makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.
 *
 *
 * Send packet to server for realm; wait for response, retransmitting
 * as necessary.
 */

//#ifdef WIN32
///* Windows Header Files: */
//#include <Winsock2.h>
//#include <Mswsock.h>

/* C RunTime Header Files */
//#include <stdlib.h>
//#include <malloc.h>
//#include <memory.h>
//#include <stdio.h>
//#include <stdlib.h>
//#else

//#ifdef VXWORKS
//#include <bcmos.h>
//#include <socket.h>
//#include <socklib.h>
//#endif

//#ifdef PSOS
//#include <pna.h>
//#include <rescfg.h>
//#endif

//#endif

#include <krb5client.h>
#include <hausenv.h>
#include <net.h>
#include "seckrb.h"

#ifdef TARGETOS_vxWorks
#include <vxWorks.h>
#include <bcmos.h>
#include <socket.h>
#include <socklib.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <time.h>
#endif

#include "KerbStatus.h"
//#include <netinet/in.h>  
/*
 * send the formatted request 'message' to a server for realm 'realm' and
 * return the response (if any) in 'reply'.
 *
 * If the message is sent and a response is received, 0 is returned,
 * otherwise an error code is returned.
 *
 * The storage for 'reply' is allocated and should be freed by the caller
 * when finished.
 */

#define krb5_max_dgram_size 5120
#define timeout_shift       15
#define SOCKET_EINTR		    EINTR
#define KRB5_SRV_UNREACH    (-1765328228L)
#ifndef WIN32
    #define INVALID_SOCKET      -1
#endif
/* MS Transport Provider IOCTL to control
 reporting PORT_UNREACHABLE messages
 on UDP sockets via recv/WSARecv/etc.
 Path TRUE in input buffer to enable (default if supported),
 FALSE to disable. */
#define SIO_UDP_CONNRESET _WSAIOW(IOC_VENDOR,12)


/********************************************************************
* Function Name:  kSetSocketIfNum()
*
* Description:  Sets the global kBcmInterfaceSoOption variable which
*               is used when creating sockets for Kerberos.
* Returns:               
*                           
*********************************************************************/
static unsigned short gKerbBcmInterfaceSoOption = 0x0c00;//BCM2_IFFLAG;;
void kSetSocketIfNum(unsigned short kBcmInterfaceSoOption)
{
    gKerbBcmInterfaceSoOption = kBcmInterfaceSoOption;
}

/********************************************************************
* Function Name:  krb5_getSocket()
*
* Description:  Creates a socket on the CableHome interface.
* Returns    :  socket descripto or zero on fail              
*                           
*********************************************************************/
int krb5_getSocket()
{
    //unsigned short bcmInterfaceSoOption = 0x0c00;//BCM2_IFFLAG;
    int optValue = 1;
    int sd;

    //open a socket
    if ( (sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1 )
    {
        LIBDEBUG("krb5_getSocket: Can't get a socket\n");
        return 0;
    }
    else
    {
        LIBDEBUG("krb5_getSocket: Opened pSocket: socket = %d\n", sd);
    }

    //set interface opt
    if ( setsockopt(sd, SOL_SOCKET, gKerbBcmInterfaceSoOption,(char *)&optValue,sizeof (optValue)) != 0 )
    {
        LIBDEBUG("krb5_getSocket: -----setsockopt failure on gKerbBcmInterfaceSoOption\n");
        return 0;
    }
    //else
    //printf("krb5_getSocket: ------setsockopt on gKerbBcmInterfaceSoOption is ok\n");

    return sd;
}
int krb5_bindSocket(int Socket, int port)
{
    struct sockaddr_in sAddr;
    //if ( ktracelevel > KTRACE_INFO )
    //    printf("krb5_bindSocket: START\n");
    
    // zero contents of target.
    memset( &sAddr, 0, sizeof(sAddr) );
    
    sAddr.sin_family = AF_INET;
    sAddr.sin_port = htons(port);
    sAddr.sin_addr.s_addr = htonl(INADDR_ANY);

    LIBDEBUG("TRYING TO BIND socket<%d> to port<%d> \n",Socket, sAddr.sin_port);
    if ( bind(Socket, (struct sockaddr *)&sAddr, sizeof(sAddr)) == -1 )
        printf("krb5_bindSocket: ERROR - we had a problem binding our trap socket errorno = 0x%x\n",errno);

    //if ( ktracelevel > KTRACE_INFO )
    //    printf("krb5_bindSocket: END\n");

    return 0;
}





krb5_error_code krb5_sendto_server(
                                  const krb5_data    *packet,
                                  const char         *server_fqdn,
                                  const int          server_port,
                                  int                krbto,
                                  krb5_data          *reply,
                                  unsigned long      reason)
{
    int             nready;
    krb5_error_code retval;
    fd_set          readable;
    struct timeval  waitlen;
    int             cc;
    int             tgssocket=0;
    int             repport;
    unsigned long   repaddr;
    int             rc;
    IPADDR          ipaddr = 0;
    int             bufsize = krb5_max_dgram_size;
//    unsigned char myip[] = {0x0A, 0x18, 0x07, 0x8E};
//    unsigned char * myIpAddr = &myip[0];

    if ( ktracelevel > KTRACE_INFO )
    {
        if ( server_fqdn )
            printf("krb5_sendto_server: server_fqdn = %s\n",server_fqdn);
        else
            printf("krb5_sendto_server: server_fqdn is NULL\n");
    }
    if ( !(reply->data = (unsigned char *)krbAllocBuf(krb5_max_dgram_size)) )
    {
        return ENOMEM;
    }

    reply->length = krb5_max_dgram_size;

    //trc:
    bcmNetIpAddrToNum(server_fqdn, &ipaddr);
    if ( ! ipaddr )
    {
        LIBDEBUG("sendto_server: bcmNetIpAddrToNum ret error\n");
        return 1;
    }


#if 0 //ORIG_WAY

    /* Open the socket */
    printf("krb5_sendto_server: trying to open socket, fqdn=%s, port=%d\n",server_fqdn,server_port);
    if ( bcmNetUdpOpen( (UINT16)server_port, TRUE, &tgssocket ) != NET_STATUS_OK )
    {
        printf("could not open socket, fqdn=%s, port=%d\n",
               server_fqdn,server_port);
        return INVALID_SOCKET;
    }
    else
    {
        printf("Socket opened successfully.\n");
    }
    //trc- not setting buf size for now
    //printf("trc- DID NOT SET BUF SIZE\n");
    /* increase size of sending buffer */
    //printf("buf size change to 0x%x, varsize = %d\n", bufsize, sizeof(bufsize));
    //if (setsockopt(tgssocket, SOL_SOCKET, SO_SNDBUF, (char *) &bufsize,
    //         sizeof(bufsize)) )
    //{
    //   printf("could not set sendbuf size, errno = 0x%x\n", errno);
    //}

    /* increase size of receiving buffer */
//   if (setsockopt(tgssocket, SOL_SOCKET, SO_RCVBUF, (char *) &bufsize,
//            sizeof(bufsize)) )
//   {
//      printf("could not set recvbuf size, errno = 0x%x\n", errno);
//   }
#endif //ORIG_WAY


    //new way
    tgssocket = krb5_getSocket();
    
    //do a bind cause provserv will send to 1293 regardless of where it came from
    //PKT-SP-SEC-I05-020116
    krb5_bindSocket(tgssocket,server_port);

    bcmNetUdpSend( tgssocket, packet->data, packet->length, ipaddr,
                   (UINT16)server_port, &rc );
    LIBDEBUG("sent to server rc = %d, tgssocket = %d\n", rc, tgssocket);


    /* if this is a response to a rekey, we do not expect a reply for AP */
    /* if this is a SA recovered message, we don't expect a reply */
    if ( ( reason == KRB_REKEY ) || ( reason == KRB_SARECOVERED) )
    {
        retval = 0;
        bcmNetUdpClose( tgssocket );
        return retval;
    }

    waitlen.tv_usec = 0;
    waitlen.tv_sec = krbto;
    FD_ZERO(&readable);
    FD_SET(tgssocket, &readable);
    if ( nready = select(FD_SETSIZE, &readable, (fd_set *)0, (fd_set *)0,
                         &waitlen) )
    {
        //printf("thread select.\n");
        if ( nready <= 0 )
        {
            LIBDEBUG("nready=%d, errno=%d\n", nready, errno);
            bcmNetUdpClose( tgssocket );
            retval = errno;
            return retval;
            /* continue; */
        }
        if ( FD_ISSET(tgssocket, &readable) )
        {
            if ( ( bcmNetUdpRecv( tgssocket, reply->data, reply->length,
                                  &repaddr,
                                  (unsigned short *)(&repport), &cc )) != NET_STATUS_OK )
            {
                LIBDEBUG("udpRecv failed.Error = %d\n", errno);
                /* man page says error could be
                 * EBADF: won't happen
                 * ENOTSOCK: it's a socket.
                 * EWOULDBLOCK: not marked non-blocking, and we selected.
                 * EINTR: could happen
                 * EFAULT: we allocated the reply packet.
                 *
                 * In addition, net related errors like ECONNREFUSED
                 * are possble (but undocumented).  Assume anything
                 * other than EINTR is a permanent error for the
                 * server (i.e. don't set sent = 1).
                 */

                // if (SOCKET_ERRNO == SOCKET_EINTR)
                /* continue; */
            }
        }

        //if ( ktracelevel > 4 )
        //    printf("udpRecv succeeded, %d bytes.\n", cc);
        reply->length = cc;
        retval = 0;
        bcmNetUdpClose( tgssocket );
        return retval;
    }
    else
    {
        LIBDEBUG("udpRecv Timeout. No reply received.\n");
        LIBDEBUG("nready=%d, errno=%d\n", nready, errno);
        
    }

    retval = KRB5_SRV_UNREACH;
    krbFreeBuf(reply->data);
    reply->length = 0;
    bcmNetUdpClose( tgssocket );
    return retval;
}
