/***************************************************************************
 *     Copyright 2001  Broadcom Corporation
 *     All Rights Reserved
 *     No portions of this material may be reproduced in any form without the
 *     written permission of:
 *              Broadcom Corporation
 *              16251 Laguna Canyon Road
 *              Irvine, California  92618
 *     All information contained in this document is Broadcom Corporation
 *     company private, proprietary, and trade secret.
 *
 ****************************************************************************
 *
 *     Filename:  timeTable.c
 *     Creation Date: 30 October 2001 (v0.00)
 *     VSS Info:
 *         $Revision: 1.2 $
 *         $Date: 2004/03/04 15:07:40 $
 *
 ****************************************************************************
 *     Description:
 *
 *       Implementation of time management for the ticket,
 *       session key and subkey
 *
 *     Notes:
 *
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <hausenv.h>
#ifdef PSOS
#include <MibAccessSec.h>
#endif
#include "timeTable.h"
#include "krbTasks.h"

#ifdef DEMO
#define GLOBAL_CFG_MAX_ENDPT   8
#else
#include <globalCfg.h>
#include <pktcMib.h>
#endif


#define MAX_TIME_TABLE_ENTRIES    GLOBAL_CFG_MAX_ENDPT

static void secPrintTimeTable(void);
static void secInitTimeTable(void);
/* structs using in table to map info to NVRAM */
typedef struct
{
   int             time_left;
   int             grace_period;
   unsigned char   fqdn[256];
   unsigned char   realm[256];
} timeTable_Tkt_Info;
typedef struct
{
   int             time_left;
   int             grace_period;
   unsigned char   fqdn[256];
} timeTable_SubKey_Info;

typedef struct
{
   int             time_left;
   int             grace_period;
} prov_tkt_time_Info;

static prov_tkt_time_Info gPROV_tkt_time;
static timeTable_Tkt_Info         gCMS_tkt_time[MAX_TIME_TABLE_ENTRIES];
static timeTable_SubKey_Info      gCMS_subkey_time[MAX_TIME_TABLE_ENTRIES];
static                            TimeTableInitialized;



#ifdef PSOS
/*
******************************************************************************
** FUNCTION:  secCleanupTimeTable
**
** PURPOSE:   Delete entry if an endpoint is provisioned to a new call agent
**
** PARAMETERS:
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
void secCleanupTimeTable(void)
{
   int             i;
   int             j;
   int             tmp_len;
   unsigned long   mib_realm_len;
   unsigned char   tmpfqdn[256];
   unsigned char   *mib_realm;
   char            *p;
   unsigned char   tmprealm[256];
   BOOL            delent;

   /* clean up subkey table */
   for ( j=0;j<MAX_TIME_TABLE_ENTRIES;j++ )
   {
      /* if entry is empty, go to next one */
      if ( ! strlen(gCMS_subkey_time[j].fqdn))
         continue;
      
      delent = TRUE;
      for ( i=0; i<GLOBAL_CFG_MAX_ENDPT; i++ )
      {
         /* get the CMS FQDN for this endpoint */
         mibAccessNcs( i, MIB_NCSITEMID_CaID, &tmp_len,
               tmpfqdn, MIB_ACCESS_GET);
         if ( ! strlen(tmpfqdn))
            continue;
         if ((p = strstr(tmpfqdn, "@")))
         {
            p++;
         }
         else
         {
            p = tmpfqdn;
         }
         while(*p)
         {
            *p++ = (char)tolower((int)*p);
         }

         if ( !memcmp(tmpfqdn, gCMS_subkey_time[j].fqdn, strlen(tmpfqdn)))
         {
            delent = FALSE;
            break;
         }
      }
      if ( delent == TRUE )
      {
         LOG(("SUBKEY TIME TABLE Entry %d call agent %s changed. Cleanup.",
                  j, gCMS_subkey_time[j].fqdn));
         tmpfqdn[strlen(gCMS_subkey_time[j].fqdn)]='\0';
         memset(gCMS_subkey_time[j].fqdn, '\0', 255);
         gCMS_subkey_time[j].time_left = 0;
         gCMS_subkey_time[j].grace_period = 0;
      }
   }

   /* clean up ticket table if any cms fqdn has changed */
   for ( j=0;j<MAX_TIME_TABLE_ENTRIES;j++ )
   {
      /* if entry is empty, go to next one */
      if ( ! strlen(gCMS_tkt_time[j].fqdn))
         continue;

      delent = TRUE;
      for ( i=0; i<GLOBAL_CFG_MAX_ENDPT; i++ )
      {
         /* get the CMS FQDN for this endpoint */
         mibAccessNcs( i, MIB_NCSITEMID_CaID, &tmp_len,
                  tmpfqdn, MIB_ACCESS_GET);
         if ( ! strlen(tmpfqdn) )
            continue;

         if ((p = strstr(tmpfqdn, "@")))
         {
            p++;
         }
         else
         {
            p = tmpfqdn;
         }
         while(*p)
         {
            *p++ = (char)tolower((int)*p);
         }

         if ( !memcmp(tmpfqdn, gCMS_tkt_time[j].fqdn, strlen(tmpfqdn)))
         {
         delent = FALSE;
            break;
         }
      }
      if ( delent == TRUE )
      {
         LOG(("TICKET TIME TABLE Entry %d (for %s)ticket fqdn changed.cleanup.", 
                  j, gCMS_tkt_time[j].fqdn));
         tmpfqdn[strlen(gCMS_tkt_time[j].fqdn)]='\0';
         memset(gCMS_tkt_time[j].fqdn, '\0', 255);
         gCMS_tkt_time[j].time_left = 0;
         gCMS_tkt_time[j].grace_period = 0;
         delent = FALSE;
      }
   }

   /* clean up ticket table if any realm has changed */
   for ( j=0;j<MAX_TIME_TABLE_ENTRIES;j++ )
   {
      /* if entry is empty, go to next one */
      if ( ! strlen(gCMS_tkt_time[j].realm))
         continue;

      delent = TRUE;
      for ( i=0; i<GLOBAL_CFG_MAX_ENDPT; i++ )
      {
         /* get the CMS FQDN for this endpoint */
         mibAccessNcs( i, MIB_NCSITEMID_CaID, &tmp_len,
               tmpfqdn, MIB_ACCESS_GET);
         if ( ! strlen(tmpfqdn) )
            continue;

         if ((p = strstr(tmpfqdn, "@")))
         {
            p++;
         }
         else
         {
            p = tmpfqdn;
         }
         while(*p)
         {
            *p++ = (char)tolower((int)*p);
         }

         /* Get the realm for this CMS from MIB */
         if (mibAccessSec (
               MIBITEM(SEC, cmsRealm),
               NULL,
               p,
               NULL,
               &mib_realm_len,
               (void*)&mib_realm,
               MIB_ITEM_GET))
         {
            LOG(("ERROR: Internal Inconsistency: Realm Name for the CMS FQDN=\"%s\" cannot be found.", p));
            LOG(("\t===>Delete This Entry from Time Table!", p));
            delent = TRUE;
            break;
         }
         /*mib_realm is a pointer into the MIB.  Make a copy so the value in the database is 
           not changed.*/
         strcpy(tmprealm, mib_realm);
         /* case insensitive */
         p = tmprealm;
         while(*p)
         {
            *p++ = (char)tolower((int)*p);
         }

         if ( !memcmp(tmprealm, gCMS_tkt_time[j].realm, mib_realm_len))
         {
            delent = FALSE;
            break;
         }
      }
      if ( delent == TRUE )
      {
         LOG(("TICKET TIME TABLE Entry %d (for %s)ticket realm changed.cleanup.
                  mib_realm = %s, table_realm = %s", 
                  j, gCMS_tkt_time[j].fqdn, tmprealm, gCMS_tkt_time[j].realm));
         tmpfqdn[strlen(gCMS_tkt_time[j].fqdn)]='\0';
         memset(gCMS_tkt_time[j].fqdn, '\0', 255);
         memset(gCMS_tkt_time[j].realm, '\0', 255);
         gCMS_tkt_time[j].time_left = 0;
         gCMS_tkt_time[j].grace_period = 0;
         delent = FALSE;
      }
   }
}
#endif

/*
******************************************************************************
** FUNCTION:  secStoreToTimeTable
**
** PURPOSE:   Store info for ticket or subkey to keep track of its expiry
**
** PARAMETERS:   flag - AS or AP, prove or CMS
**               cmsfqdn - fqdn of cms
**               realm - realm
**               credentials - ticket and session key
**               appkey  -  subkey
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
void secStoreToTimeTable(
      SEC_MSG_FLAG         flag,
      unsigned char        *cmsfqdn,
      unsigned char        *realm,
      MtaCredentials       *credentials,
      MtaApplicationKey    *appkey)
{
   int              i;
   int              gp;
   int              tmp;
   unsigned char    lcmsfqdn[255];
   unsigned char    lrealm[255];
   unsigned char    *p;


   secInitTimeTable();

   strcpy(lcmsfqdn, cmsfqdn);
   p = lcmsfqdn;
   while(*p)
   {
      *p++ = (char)tolower((int)*p);
   }
   strcpy(lrealm, realm);
   p = lrealm;
   while(*p)
   {
      *p++ = (char)tolower((int)*p);
   }

#ifdef PSOS
   /* remove entry if call agent is changed */
   secCleanupTimeTable();
#endif

   switch (flag)
   {
      case AS_PROV_FLAG:
         /* provisioning ticket: store to entry 0 */
         gPROV_tkt_time.time_left = credentials->tkt_times.end_time -
            credentials->tkt_times.auth_time;
         /* get grace period from MIB */
#ifdef PSOS
         if ( mibAccessSec(
               MIBITEM(SEC, realmPkinitGracePeriod),
               NULL,
               realm,
               NULL,
               &tmp,
               (void*)&gp,
               MIB_ITEM_GET))
         {
            LOG(("ERROR: StoreTimeTable no MIB grace period for realm %s, setting to default %d.", realm, KRB_DEFAULT_PKINIT_GRACE_PERIOD));
         }
#endif
         if ( gp <= 0 )
            gPROV_tkt_time.grace_period = KRB_DEFAULT_PKINIT_GRACE_PERIOD;
         else
            gPROV_tkt_time.grace_period = gp * 60;
         break;

      case AS_CMS_FLAG:
#ifdef PSOS
         /* Get the grace period from MIB */
         if ( mibAccessSec(
               MIBITEM(SEC, realmPkinitGracePeriod),
               NULL,
               realm,
               NULL,
               &tmp,
               (void*)&gp,
               MIB_ITEM_GET))
         {
            LOG(("ERROR: StoreTimeTable no MIB grace period for realm %s, setting to default %d.", realm, KRB_DEFAULT_PKINIT_GRACE_PERIOD));
         }
#endif
         if ( gp <= 0 )
            gp = KRB_DEFAULT_PKINIT_GRACE_PERIOD;
         else
            gp *= 60;

         for ( i=0;i<MAX_TIME_TABLE_ENTRIES;i++ )
         {
            /* if there is an existing entry, overwrite the values */
            if ( !memcmp(lcmsfqdn, gCMS_tkt_time[i].fqdn,strlen(lcmsfqdn)))
            {
               strcpy(gCMS_tkt_time[i].realm, lrealm);
               gCMS_tkt_time[i].time_left = credentials->tkt_times.end_time -
                  credentials->tkt_times.auth_time;
               gCMS_tkt_time[i].grace_period = gp;
               break;
            }
         }

         /* if no entries matched, this is a new entry */
         if ( i == MAX_TIME_TABLE_ENTRIES )
         {
            for ( i=0;i<MAX_TIME_TABLE_ENTRIES;i++ )
            {
               if ( gCMS_subkey_time[i].fqdn[0] == '\0')
               {
                  strcpy(gCMS_tkt_time[i].fqdn, lcmsfqdn);
                  strcpy(gCMS_tkt_time[i].realm, lrealm);
                  gCMS_tkt_time[i].time_left = credentials->tkt_times.end_time -
                     credentials->tkt_times.auth_time;
                  gCMS_tkt_time[i].grace_period = gp;
                  break;
               }
            }
         }
         if ( i==MAX_TIME_TABLE_ENTRIES )
            LOG(("TIMETABLE PANIC: NO MORE ENTRIES LEFT FOR CMS SUBKEYS."));
         break;

      case AP_CMS_FLAG:
         for ( i=0;i<MAX_TIME_TABLE_ENTRIES;i++ )
         {
            /* if there is an existing entry, overwrite the values */
            if ( !memcmp(lcmsfqdn, gCMS_subkey_time[i].fqdn,strlen(lcmsfqdn)))
            {
               gCMS_subkey_time[i].time_left = appkey->sec_params_lifetime;
               gCMS_subkey_time[i].grace_period = appkey->grace_period;
               break;
            }
         }

         /* if no entries matched, this is a new entry */
         if ( i == MAX_TIME_TABLE_ENTRIES )
         {
            for ( i=0;i<MAX_TIME_TABLE_ENTRIES;i++ )
            {
               if ( gCMS_subkey_time[i].fqdn[0] == '\0')
               {
                  strcpy(gCMS_subkey_time[i].fqdn, lcmsfqdn);
                  gCMS_subkey_time[i].time_left = appkey->sec_params_lifetime;
                  gCMS_subkey_time[i].grace_period = appkey->grace_period;
                  break;
               }
            }
         }
         if ( i==MAX_TIME_TABLE_ENTRIES )
            LOG(("TIMETABLE PANIC: NO MORE ENTRIES LEFT FOR CMS SUBKEYS."));
         break;

      default:
         LOG(("KERBEROS TIME TABLE ERROR:secStoreToTimeTable bad flag(%d).",
                  flag));
   }
   secPrintTimeTable();
}


/*
******************************************************************************
** FUNCTION:  secUpdateTimeTable
**
** PURPOSE:   renew any ticket or subkey if expired
**
** PARAMETERS:   the time between sleeps for the krb task
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
void secUpdateTimeTable(unsigned short sleep_time)
{
   int             flag=-1;
   int             i;
   unsigned char   tmpfqdn[255];
   unsigned char   *p;

   /* for provisioning ticket */
   if ( gPROV_tkt_time.time_left > 0 )
   {
      gPROV_tkt_time.time_left -= sleep_time;
   }
   if ( ( gPROV_tkt_time.time_left <= gPROV_tkt_time.grace_period ) &&
         ( gPROV_tkt_time.grace_period ))
   {
      flag = AS_PROV_FLAG;
      /* No need to pass fqdn here, pass anything */
      LOG(("PROV Ticket Expired, Renew ..."));
      reAuthenticate(flag, tmpfqdn);
   }

   /* for CMS ticket */
   for (i=0;i<MAX_TIME_TABLE_ENTRIES;i++)
   {
      if ( gCMS_tkt_time[i].time_left > 0 )
      {
         gCMS_tkt_time[i].time_left -= sleep_time;
      }
      if ( ( gCMS_tkt_time[i].time_left <= gCMS_tkt_time[i].grace_period ) &&
           ( gCMS_tkt_time[i].grace_period ))
      {
         flag = AS_CMS_FLAG;
         strcpy(tmpfqdn, gCMS_tkt_time[i].fqdn);
         p = tmpfqdn;
         while(*p)
         {
            *p++ = (char)tolower((int)*p);
         }
         LOG(("CMS Ticket Expired, Renew for %s ...", tmpfqdn));
         reAuthenticate(flag, tmpfqdn);
      }
   }

   /* for CMS subkey */
   for (i=0;i<MAX_TIME_TABLE_ENTRIES;i++)
   {
      if ( gCMS_subkey_time[i].time_left > 0 )
      {
         gCMS_subkey_time[i].time_left -= sleep_time;
      }
      if ( ( gCMS_subkey_time[i].time_left <= gCMS_subkey_time[i].grace_period ) &&
           ( gCMS_subkey_time[i].grace_period ))
      {
         flag = AP_CMS_FLAG;
         strcpy(tmpfqdn, gCMS_subkey_time[i].fqdn);
         p = tmpfqdn;
         while(*p)
         {
            *p++ = (char)tolower((int)*p);
         }
         LOG(("CMS Subkey Expired, Renew for %s ...", tmpfqdn));
         reAuthenticate(flag, tmpfqdn);
      }
   }   
}


/*
 ******************************************************************************
 ** FUNCTION:  secPrintTimeTable
 **
 ** PURPOSE:   
 **
 ** PARAMETERS:   
 **
 ** RETURNS:
 **
 ** NOTE:
 **
 ******************************************************************************
 */
void secPrintTimeTable(void)
{
   int   i;
   //trc-
   return;

   printf("CMS Ticket Time Table:\n");
   for ( i=0;i<MAX_TIME_TABLE_ENTRIES;i++ )
   {
      printf("entry %d, fqdn = %s, realm = %s, time_left = %d, grace_period = %d.\n",
            i, gCMS_tkt_time[i].fqdn, gCMS_tkt_time[i].realm, 
            gCMS_tkt_time[i].time_left, gCMS_tkt_time[i].grace_period);
   }
   printf("\nCMS SubKey Time Table:\n");
   for ( i=0;i<MAX_TIME_TABLE_ENTRIES;i++ )
   {
      printf("entry %d, fqdn = %s, time_left = %d, grace_period = %d.\n",
            i, gCMS_subkey_time[i].fqdn, 
            gCMS_subkey_time[i].time_left, gCMS_subkey_time[i].grace_period);
   }
}

/*
 ******************************************************************************
 ** FUNCTION:  secInitTimeTable
 **
 ** PURPOSE:   
 **
 ** PARAMETERS:   
 **
 ** RETURNS:
 **
 ** NOTE:
 **
 ******************************************************************************
 */
void secInitTimeTable(void)
{
   if ( ! TimeTableInitialized )
   {
      printf("Initializing Time Tables (TimeTableInitialized=%d).\n", TimeTableInitialized);
      TimeTableInitialized = 1;
      memset(gCMS_tkt_time, 0, sizeof(gCMS_tkt_time));
      memset(gCMS_subkey_time, 0, sizeof(gCMS_subkey_time));
   }
}


