/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: socket.h
*    Creation Date: 28 June 2001 (v0.00)
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:54:42 $
*
****************************************************************************
*    Description:
*
*      Private socket related prototypes and type defines for VxWorks.
*
****************************************************************************/

#ifndef _SOCKET_H
#define _SOCKET_H

#ifdef __cplusplus
extern "C" {
#endif
#include "bcmosHelper.h"
/*
*****************************************************************************
** TYPE DEFINITIONS
*****************************************************************************
*/

/* OS-specific status types. This is the data type returned by (most) VxWorks
 * socket related API calls */
typedef int       NATIVE_SOCKET_STATUS;

/*
*****************************************************************************
** FUNCTION PROTOTYPES
*****************************************************************************
*/

/*****************************************************************************
** FUNCTION:   checkSocketStatus
**
** PURPOSE:    This function takes the status returned by an OS-specific socket API
**             function call. It maps this to a generic BcmOS return status, and
**             logs any errors.
**
** PARAMETERS: status   - operating system specific status returned from system call
**             fnc      - name of calling function, used for error LOG statements
**
** RETURNS:    Operating system independent status value of type OS_STATUS
**             (that corresponds to 'status') . BcmOS public interface functions
**             should return OS independent status values. Native OS status
**             values should be encapsulated; however, they may be LOGed for
**             debug purposes.
**
** NOTES:      This function is used to validate all OS-specific socket-related API
**             calls invoked by the BcmOS layer. However, not all OS API calls return
**             consistent status values/error codes. Therefore, the following assumptions
**             have been made:
**
**             PSOS:    an error is indicated by a 'status' value of -1
**             VXWORKS: an error is indicated by a 'status' value of ERROR
**             WIN32:   an error is indicated by a 'status' value of INVALID_SOCKET
**                      or SOCKET_ERROR
**
**             These assumptions should handle most OS-specific API calls. However, it
**             is the responsibility of the caller to massage the input parameters to conform
**             to these assumptions for any OS API calls that do not follow these rules.
**
*****************************************************************************/
OS_STATUS checkSocketStatus( NATIVE_SOCKET_STATUS status, char *fnc );



/*****************************************************************************
** FUNCTION:   socketInit
**
** PURPOSE:    Initialize any socket related interfaces
**
** PARAMETERS: None
**
** RETURNS:    OS_STATUS_OK or OS_STATUS_ERR
*****************************************************************************/
OS_STATUS socketInit( void );


/*****************************************************************************
** FUNCTION:   socketSetInit
**
** PURPOSE:    Initialize any socket set related interfaces
**
** PARAMETERS: None
**
** RETURNS:    OS_STATUS_OK or OS_STATUS_ERR
*****************************************************************************/
OS_STATUS socketSetInit( void );


#ifdef __cplusplus
    }
#endif

#endif  /* _SOCKET_H  */
