/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: systime.c
*    Creation Date: 27 June 2001 (v0.00)
*    VSS Info:
*        $Revision: 1.2 $
*        $Date: 2002/12/23 15:13:28 $
*
****************************************************************************
*    Description:
*
*     Broadcom Generic Operating System Functions. This file contains the VxWorks
*     specific implementation of system time management.
*
****************************************************************************/
     
///#include <hausenv.h>
///#include <string.h>
///#include <bcmos.h>

/* Operating system specific includes */
#include <vxWorks.h>
#include <tickLib.h>
#include "vxwrkdef.h"


#if CH_NEEDED
/*****************************************************************************
** FUNCTION:   bcmOsGetTime
**
** PURPOSE:    get current system time in ticks.
**
** PARAMETERS: osTime - pointer to OS_TICKS (output)
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       This function is only used to measure relative time for the
**             purposes of timing out syncronous requests
*****************************************************************************/
OS_STATUS bcmOsGetTime(OS_TICKS *osTime)
{
   ASSERT( osTime );

   *osTime = tickGet();

   /* Convert to OS-specific tick value */
   *osTime *= OS_TICK_SCALE_FACTOR;

   /* Error checking */
   return ( OS_STATUS_OK );
}
#endif

/*****************************************************************************
** FUNCTION:   bcmOsGetDeltaTime
**
** PURPOSE:    Get delta system time in ticks.
**
** PARAMETERS: prevTime - previous OS_TICKS value (input)
**             curTime - current OS_TICKS value (output)
**             deltaTime - ticks since 'prevTime' (output)
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       This function is only used to measure relative time for the
**             purposes of timing out syncronous requests. It will handle
**             wrap around conditions.
*****************************************************************************/
#if CH_NEEDED
OS_STATUS bcmOsGetDeltaTime(OS_TICKS prevTime, OS_TICKS *curTime, OS_TICKS *deltaTime)
{
   OS_STATUS   status;

   ASSERT( curTime && deltaTime );

   if (( status = bcmOsGetTime( curTime )) == OS_STATUS_OK )
   {
      /* Calculate delta time - take care of the wrap around case */
      if ( *curTime < prevTime )
      {
         *deltaTime = OS_MAX_TIME - prevTime + *curTime + 1;
      }
      else
      {
         *deltaTime = *curTime - prevTime;
      }
   }

   return ( status );
}
#endif

/*****************************************************************************
** FUNCTION:   bcmOsGetTimeEx
**
** PURPOSE:    gets full time related information (date, time, ticks).
**
** PARAMETERS: date  - returns the Date related bits (y, m, d)
**             time  - returns the Time related bits (h, m, s)
**             time  - returns ticks
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       VxWorks not supported
**
*****************************************************************************/
OS_STATUS bcmOsGetTimeEx(OS_TMDATE *date, OS_TMTIME *time, OS_TICKS *ticks)
{
   NATIVE_STATUS status = ERROR;

   /* Convert to OS-specific tick value */
   *ticks *= OS_TICK_SCALE_FACTOR;

   /* Error checking */
   return ( checkStatus( status, "bcmOsGetTimeEx" ));
}


/*****************************************************************************
** FUNCTION:   bcmOsSetTime
**
** PURPOSE:    Sets full time related information (date, time, ticks).
**
** PARAMETERS: date  - sets the Date related bits (y, m, d)
**             time  - sets the Time related bits (h, m, s)
**             time  - sets ticks
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:       VxWorks not supported
**
*****************************************************************************/
#if CH_NEEDED
OS_STATUS bcmOsSetTime(OS_TMDATE date, OS_TMTIME time, OS_TICKS ticks)
{
   NATIVE_STATUS  status = ERROR;

   /* Convert to OS-specific tick value */
   ticks = ticks / OS_TICK_SCALE_FACTOR;

   /* Error checking */
   return( checkStatus( status, "bcmOsSetTime" ) );
}
#endif
