/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16215 Alton Parkway
*             P.O. Box 57013
*             Irvine, California 92619-7013
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: krbbuf.c
*
****************************************************************************
*    Description:
*
*      This file contains functions to manage kerberos memory.
*
****************************************************************************/

/*
*****************************************************************************
** INCLUDE FILES
*****************************************************************************
*/
#include <stdio.h>
/* #include <cxcLog.h> */
/* #include <secLog.h> */
#include <seckrb.h>
#include "krbbuf.h"
#include "ParseLibMain.h"

int alloc_count_xl;
int alloc_count_l;
int alloc_count_m;
int alloc_count_s;
int free_count_xl;
int free_count_l;
int free_count_m;
int free_count_s;


#define krbmaxbufpool_xlarge    6
#define krbmaxbufpool_large     30
#define krbmaxbufpool_medium    40
#define krbmaxbufpool_small     164

/*
*****************************************************************************
** TYPE DEFINITIONS and CONSTANTS
*****************************************************************************
*/
typedef struct krbbuf_xlarge
{
   char             data[krbmaxbuflen_xlarge];
   unsigned long    bsize;
   struct           krbbuf_xlarge *next;
} krbbuf_xlarge;
typedef struct krbbuf_large
{
   char             data[krbmaxbuflen_large];
   unsigned long    bsize;
   struct           krbbuf_large *next;
} krbbuf_large;

typedef struct krbbuf_medium
{
   char             data[krbmaxbuflen_medium];
   unsigned long    bsize;
   struct           krbbuf_medium *next;
} krbbuf_medium;

typedef struct krbbuf_small
{
   char             data[krbmaxbuflen_small];
   unsigned long    bsize;
   struct           krbbuf_small *next;
} krbbuf_small;
/*
*****************************************************************************
** PRIVATE VARIABLES
*****************************************************************************
*/
typedef struct _krbmeminfo
{
   krbbuf_small *freechain_small;           /* Next free buffer */
   krbbuf_small krbBufferPool_small[krbmaxbufpool_small];

   krbbuf_medium *freechain_medium;           /* Next free buffer */
   krbbuf_medium krbBufferPool_medium[krbmaxbufpool_medium];  

   krbbuf_large *freechain_large;           /* Next free buffer */
   krbbuf_large krbBufferPool_large[krbmaxbufpool_large];

   krbbuf_xlarge *freechain_xlarge;           /* Next free buffer */
   krbbuf_xlarge krbBufferPool_xlarge[krbmaxbufpool_xlarge];
} krbmeminfo;

static krbmeminfo krbminfo ;

/*
*****************************************************************************
** FUNCTION:   krbAllocBuf_large
**
** PURPOSE:    Allocate buffer from pool
**
** PARAMETERS: none
**
** RETURNS:    char *newbuffer or NULL if no more buffers
**
*****************************************************************************
*/
char *krbAllocBuf_large(unsigned long ksize)
{
   krbbuf_large *p = krbminfo.freechain_large;

   // printf("\t **** krbAllocBuf_large.\n");
   if (p)
   {
      krbminfo.freechain_large = p->next;
   }
   else
   {
      if (gprintFlag)
	  {
         printf("ERROR: no more free large buffers\n");
	  }
      return NULL;
   }
   p->bsize = ksize;
   alloc_count_l++;
   return (char *) p;
}


/*
*****************************************************************************
** FUNCTION:   krbAllocBuf_large
**
** PURPOSE:    Allocate buffer from pool
**
** PARAMETERS: none
**
** RETURNS:    char *newbuffer or NULL if no more buffers
**
*****************************************************************************
*/
char *krbAllocBuf_xlarge(unsigned long ksize)
{
   krbbuf_xlarge *p = krbminfo.freechain_xlarge;

   // printf("\t **** krbAllocBuf_large.\n");
   if (p)
   {
      krbminfo.freechain_xlarge = p->next;
   }
   else
   {
      if (gprintFlag)
	  {
         printf("ERROR: no more free X-large buffers\n");
	  }
      return NULL;
   }
   p->bsize = ksize;
   alloc_count_xl++;
   return (char *) p;
}


/*
*****************************************************************************
** FUNCTION:   krbFreeBuf_xlarge
**
** PURPOSE:    Release buffer back to pool
**
** PARAMETERS: buf - buffer to free
**
** RETURNS:    none
**
*****************************************************************************
*/
void krbFreeBuf_xlarge(char *buf)
{
   krbbuf_xlarge *p = (krbbuf_xlarge *) buf;
   // printf("\t *** krbFreeBuf_xlarge.\n");

   if (p < &(krbminfo.krbBufferPool_xlarge[0]) || 
         p > &(krbminfo.krbBufferPool_xlarge[krbmaxbufpool_xlarge-1]))
   {
      if (gprintFlag)
	  {
         printf("ERROR: Freeing invalid large buffer pointer\n");
      }
   }
   else
   {
      p->next = krbminfo.freechain_xlarge;
      p->bsize = 0;
      krbminfo.freechain_xlarge = p;
   }
   free_count_xl++;
}


/*
*****************************************************************************
** FUNCTION:   krbFreeBuf_large
**
** PURPOSE:    Release buffer back to pool
**
** PARAMETERS: buf - buffer to free
**
** RETURNS:    none
**
*****************************************************************************
*/
void krbFreeBuf_large(char *buf)
{
   krbbuf_large *p = (krbbuf_large *) buf;
   // printf("\t *** krbFreeBuf_large.\n");

   if (p < &(krbminfo.krbBufferPool_large[0]) || 
         p > &(krbminfo.krbBufferPool_large[krbmaxbufpool_large-1]))
   {
      if (gprintFlag)
	  {
         printf("ERROR: Freeing invalid large buffer pointer\n");
      }
   }
   else
   {
      p->next = krbminfo.freechain_large;
      p->bsize = 0;
      krbminfo.freechain_large = p;
   }
   free_count_l++;
}

/*
*****************************************************************************
** FUNCTION:   krbInitBufs_large
**
** PURPOSE:    Initialize buffer pool
**
** PARAMETERS: none
**
** RETURNS:    none
**
*****************************************************************************
*/
void krbInitBufs_large(void)
{
   int i = krbmaxbufpool_large;

   krbminfo.freechain_large = NULL;
   do
   {
      krbFreeBuf_large((char *)&(krbminfo.krbBufferPool_large[--i]));
   }
   while (i);
}


/*
*****************************************************************************
** FUNCTION:   krbInitBufs_xlarge
**
** PURPOSE:    Initialize buffer pool
**
** PARAMETERS: none
**
** RETURNS:    none
**
*****************************************************************************
*/
void krbInitBufs_xlarge(void)
{
   int i = krbmaxbufpool_xlarge;

   krbminfo.freechain_xlarge = NULL;
   do
   {
      krbFreeBuf_xlarge((char *)&(krbminfo.krbBufferPool_xlarge[--i]));
   }
   while (i);
}




/*
*****************************************************************************
** FUNCTION:   krbAllocBuf_medium
**
** PURPOSE:    Allocate buffer from pool
**
** PARAMETERS: none
**
** RETURNS:    char *newbuffer or NULL if no more buffers
**
*****************************************************************************
*/
char *krbAllocBuf_medium(unsigned int ksize)
{
   krbbuf_medium *p = krbminfo.freechain_medium;
   // printf("\t*** krbAllocBuf_medium.\n");
   if (p)
   {
      krbminfo.freechain_medium = p->next;
   }
   else
   {
      if (gprintFlag)
	  {
         printf("ERROR: no more medium free buffers\n");
	  }
      return NULL;
   }
   p->bsize = ksize;
   alloc_count_m++;
   return (char *) p;
}

/*
*****************************************************************************
** FUNCTION:   krbFreeBuf_medium
**
** PURPOSE:    Release buffer back to pool
**
** PARAMETERS: buf - buffer to free
**
** RETURNS:    none
**
*****************************************************************************
*/
void krbFreeBuf_medium(char *buf)
{
   krbbuf_medium *p = (krbbuf_medium *) buf;
   // printf("\t*** krbFreeBuf_medium.\n");
   if (p < &(krbminfo.krbBufferPool_medium[0]) || 
         p > &(krbminfo.krbBufferPool_medium[krbmaxbufpool_medium-1]))
   {
      if (gprintFlag)
	  {
         printf("ERROR: Freeing invalid medium buffer pointer\n");
      }
   }
   else
   {
      p->next = krbminfo.freechain_medium;
      p->bsize = 0;
      krbminfo.freechain_medium = p;
   }
   free_count_m++;
}

/*
*****************************************************************************
** FUNCTION:   krbInitBufs_medium
**
** PURPOSE:    Initialize buffer pool
**
** PARAMETERS: none
**
** RETURNS:    none
**
*****************************************************************************
*/
void krbInitBufs_medium(void)
{
   int i = krbmaxbufpool_medium;

   krbminfo.freechain_medium = NULL;
   do
   {
      krbFreeBuf_medium((char *)&(krbminfo.krbBufferPool_medium[--i]));
   }
   while (i);
}




/*
*****************************************************************************
** FUNCTION:   krbAllocBuf_small
**
** PURPOSE:    Allocate buffer from pool
**
** PARAMETERS: none
**
** RETURNS:    char *newbuffer or NULL if no more buffers
**
*****************************************************************************
*/
char *krbAllocBuf_small(unsigned int ksize)
{
   krbbuf_small *p = krbminfo.freechain_small;
   // printf("\t*** krbAllocBuf_small.\n");
   if (p)
   {
      krbminfo.freechain_small = p->next;
   }
   else
   {
      if (gprintFlag)
	  {
         printf("ERROR: no more small free buffers\n");
	  }
      return NULL;
   }
   p->bsize = ksize;
   alloc_count_s++;
   return (char *) p;
}

/*
*****************************************************************************
** FUNCTION:   krbFreeBuf_small
**
** PURPOSE:    Release buffer back to pool
**
** PARAMETERS: buf - buffer to free
**
** RETURNS:    none
**
*****************************************************************************
*/
void krbFreeBuf_small(char *buf)
{
   krbbuf_small *p = (krbbuf_small *) buf;
   // printf("\t*** krbFreeBuf_small.\n");
   if (p < &(krbminfo.krbBufferPool_small[0]) || 
         p > &(krbminfo.krbBufferPool_small[krbmaxbufpool_small-1]))
   {
      if (gprintFlag)
	  {
         printf("ERROR: Freeing invalid small buffer pointer\n");
      }
   }
   else
   {
      p->next = krbminfo.freechain_small;
      p->bsize = 0;
      krbminfo.freechain_small = p;
   }
   free_count_s++;
}

/*
*****************************************************************************
** FUNCTION:   krbInitBufs_small
**
** PURPOSE:    Initialize buffer pool
**
** PARAMETERS: none
**
** RETURNS:    none
**
*****************************************************************************
*/
void krbInitBufs_small(void)
{
   int i = krbmaxbufpool_small;

   krbminfo.freechain_small = NULL;
   do
   {
      krbFreeBuf_small((char *)&(krbminfo.krbBufferPool_small[--i]));
   }
   while (i);

}



/*
*****************************************************************************
** FUNCTION:   krbInitBufs
**
** PURPOSE:    Initialize buffer pool
**
** PARAMETERS: none
**
** RETURNS:    none
**
*****************************************************************************
*/
void krbInitBufs(void)
{
   krbInitBufs_xlarge();
   krbInitBufs_large();
   krbInitBufs_medium();
   krbInitBufs_small();
      
   alloc_count_s = 0;
   free_count_s = 0;
   alloc_count_m = 0;
   free_count_m = 0;
   alloc_count_l = 0;
   free_count_l = 0;
   alloc_count_xl = 0;
   free_count_xl = 0;

   
   /*SEC_LOG(("krbInitBufs Initializing buffers."));*/
}

/*
*****************************************************************************
** FUNCTION:    krbAllocBuf
**
** PURPOSE:     find out which buffer and allocate space for buffer
**
** PARAMETERS:  size - in: size of the buffer
**
** RETURNS:    char *newbuffer or NULL if no more buffers
**
*****************************************************************************
*/
char *krbAllocBuf(unsigned long  ksize)
{
   if ( ksize > krbmaxbuflen_xlarge )
   {
      if (gprintFlag)
	  {
         printf("ERROR: buffer size %d too large\n", ksize);
	  }
      return NULL;
   }
   
   if ( ksize >= krbmaxbuflen_large )
   {
      return (krbAllocBuf_xlarge(ksize));
   }
   
   if ( ksize >= krbmaxbuflen_medium )
   {
      return (krbAllocBuf_large(ksize));
   }
   
   if ( ksize >= krbmaxbuflen_small )
   {
      return (krbAllocBuf_medium(ksize));
   }

   return (krbAllocBuf_small(ksize));
}



/*
*****************************************************************************
** FUNCTION:    krbReallocBuf
**
** PURPOSE:     Reallocate buffer
**
** PARAMETERS:  size - in: size of the buffer
**
** RETURNS:    char *newbuffer or NULL if no more buffers
**
*****************************************************************************
*/
char *krbReallocBuf(char *inbuf, unsigned long oldsize, unsigned long  ksize)
{
   char *p;
   krbbuf_large *s = (krbbuf_large *)inbuf;
   if ( ksize > krbmaxbuflen_xlarge )
   {
      if (gprintFlag)
	  {
         printf("ERROR: (realloc) buffer size %d too large\n", ksize);
	  }
      return NULL;
   }
   p = (char *) krbAllocBuf(ksize);
   memcpy(p, inbuf, oldsize);
   krbFreeBuf(inbuf);
   return (char *) p;
}



/*
*****************************************************************************
** FUNCTION:    krbCallocBuf
**
** PURPOSE:     Alocate space enough array from buf pool
**
** PARAMETERS:  size - in: size of the buffer
**
** RETURNS:    char *newbuffer or NULL if no more buffers
**
*****************************************************************************
*/
char *krbCallocBuf(unsigned long  asize, unsigned long ksize)
{
   char *p;
   unsigned long s = asize * ksize;
   
   if ( s > krbmaxbuflen_xlarge )
   {
      if (gprintFlag)
	  {
         printf("ERROR: (calloc) buffer size %d too large\n", s);
	  }
      return NULL;
   }
   p = (char *)krbAllocBuf(s);
   if ( p == NULL )
   {
      printf("krbCallocBuf: Failed to allocate buffer buffer !!! \n");
   }
   memset(p, 0, s);
   return (p);
}


/*
*****************************************************************************
** FUNCTION:   krbFreeBuf
**
** PURPOSE:    Release buffer back to pool
**
** PARAMETERS: buf - buffer to free
**
** RETURNS:    none
**
*****************************************************************************
*/
void krbFreeBuf(char *buf)
{
   char *p = buf;
   
   if ( ((krbbuf_xlarge *)p > &(krbminfo.krbBufferPool_xlarge[krbmaxbufpool_xlarge-1])) ||
         ( (krbbuf_small *)p < &(krbminfo.krbBufferPool_small[0])))
   {
      if (gprintFlag)
	  {
         printf("ERROR: Freeing invalid buffer pointer\n");
      }
   }
   else if ( (krbbuf_small *)p <  &(krbminfo.krbBufferPool_small[krbmaxbufpool_small]))
   {
      krbFreeBuf_small(buf);
   }
   else if ( (krbbuf_medium *)p < &(krbminfo.krbBufferPool_medium[krbmaxbufpool_medium]))
   {
      krbFreeBuf_medium(buf);
   }
   else if ( (krbbuf_large *)p < &(krbminfo.krbBufferPool_large[krbmaxbufpool_large]))
   {
      krbFreeBuf_large(buf);
   }
   else if ( (krbbuf_xlarge *)p < &(krbminfo.krbBufferPool_xlarge[krbmaxbufpool_xlarge]))
   {
      krbFreeBuf_xlarge(buf);
   }
}


void krbPrintBufCounts(void)
{
	printf("small alloced %d, freed %d.\n", alloc_count_s, free_count_s);
    printf("meduim alloced %d, freed %d.\n", alloc_count_m, free_count_m);
	printf("large alloced %d, freed %d.\n", alloc_count_l, free_count_l);
	printf("Xlarge alloced %d, freed %d.\n", alloc_count_xl, free_count_xl);
}