/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: hausenv.h
*    Creation Date:
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:53:16 $
*
****************************************************************************
*    Description:
*
*      This file contains system-wide types and definitions
*
****************************************************************************/

#ifndef HAUSENV_H
#define HAUSENV_H

#ifdef __cplusplus
extern "C" {
#endif

/***********************************************
* Figure out what environment we're running in.
************************************************/
///#include <haus_compiler.h>
//#include <haus_cpu.h>
//#include <haus_os.h>

#include <limits.h>        /* C-type limit defines */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

/******************************
**  Standard variable types   *
******************************/

#if ( HAUS_OS_PSOS ) || ( HAUS_OS_WIN32 )
   typedef unsigned char   UINT8;      /* WARNING: CHAR is platform dependant */
   typedef unsigned short  UINT16;     /* UINT16 is platform independent */
   typedef unsigned int    UINT32;     /* UINT32 is platform independent */
   typedef int BOOL;
#endif

typedef signed short       SINT16;     /* SINT16 is platform independent */
typedef signed long        SINT32;     /* SINT32 is platform independent */
typedef UINT32             IPADDR;     /* IP4 Ip Address data type defined. This data type is needed all over the places! */


#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE  1
#endif


#define SUCCESS   0
#define FAILURE   1


/*******************************
** Definitions for INLINE
*******************************/
#if HAUS_COMPILER_VC

   /*
   * inline isn't supported when compiling C files, but __inline is (since
   * strictly speaking inline isn't a C feature)
   */

   #if !defined( __cplusplus )
   #define  inline   __inline
   #endif

#elif HAUS_COMPILER_DCC

   #define INLINE inline

#elif HAUS_COMPILER_GNU

   #define INLINE __inline__

#endif

/******************************
**  Standard variable limits  *
******************************/

#define MAX_UINT16      USHRT_MAX
#define MAX_SINT16      SHRT_MAX
#define MAX_UINT32      ULONG_MAX
#define MAX_SINT32      LONG_MAX

#if (MAX_UINT16 != 0xffff)
#error Maximum UINT16 != 65535
#endif

#if (MAX_SINT16 != 0x7fff)
#error Maximum SINT16 != 32767
#endif

#if (MAX_UINT32 != 0xffffffff)
#error Maximum UINT32 != 4294967295
#endif

#if (MAX_SINT32 != 0x7fffffff)
#error Maximum SINT32 != 2147483647
#endif


/***********************************
** Minimum and maximum macros      *
***********************************/
#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif


/***********************************/

/*
** Assert Macro Definition
*/

#define ASSERT(p) ((p) ? (void)0 : (void) (fprintf(stderr, \
                    "Assertion failed: %s, file %s, line %d", \
                    #p, __FILE__, __LINE__ ), abort()))

#define PANIC(m)     Log(m)
#define LOG(fmt)     Log fmt
#define LOGMSG(buf)  LogMsg buf

/***********************************
** Error logging                   *
***********************************/


extern void Log( char const *format,...);
extern void LogMsg( char *string );


/***********************************
**                                 *
***********************************/

#ifndef __sdsp__

/*
** Attribute declaration only having meaning with GNU compiler
*/
#define ATTRIBUTE_UNUSED
#define ATTRIBUTE_SECTION(sectionName)

#else

#define ATTRIBUTE_UNUSED __attribute__((unused))
#define ATTRIBUTE_SECTION(sectionName) sectionName

/*
** Assert Macro Definition
*/

#define ASSERT(expr) if (!(expr)) { asm(" bits %ireq, 15"); }

#endif

#ifdef __cplusplus
    }
#endif

#endif   /* HAUSENV_H */

